#------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------

""" Defines the BasicEditorFactory class, which allows creating editor
    factories that use the same class for creating all editor styles.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Any

from .editor_factory import EditorFactory

#-------------------------------------------------------------------------
#  'BasicEditorFactory' base class:
#-------------------------------------------------------------------------


class BasicEditorFactory(EditorFactory):
    """ Base class for editor factories that use the same class for creating
        all editor styles.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Editor class to be instantiated
    klass = Any

    #-------------------------------------------------------------------------
    #  Property getters.
    #-------------------------------------------------------------------------

    def _get_simple_editor_class(self):
        """ Returns the editor class to use for "simple" style views.
        Overridden to return the value of the 'klass' trait.

        """
        return self.klass

    def _get_custom_editor_class(self):
        """ Returns the editor class to use for "custom" style views.
        Overridden to return the value of the 'klass' trait.

        """
        return self.klass

    def _get_text_editor_class(self):
        """ Returns the editor class to use for "text" style views.
        Overridden to return the value of the 'klass' trait.

        """
        return self.klass

    def _get_readonly_editor_class(self):
        """ Returns the editor class to use for "readonly" style views.
        Overridden to return the value of the 'klass' trait.

        """
        return self.klass

    #-------------------------------------------------------------------------
    #  Allow an instance to be called:
    #-------------------------------------------------------------------------

    def __call__(self, *args, **traits):
        return self.set(**traits)

## EOF ########################################################################



from __future__ import unicode_literals
from __future__ import print_function

from fs import SEEK_SET, SEEK_CUR, SEEK_END

import io
from functools import wraps

import six


class RawWrapper(object):
    """Convert a Python 2 style file-like object in to a IO object"""
    def __init__(self, f, mode=None, name=None):
        self._f = f
        self.is_io = isinstance(f, io.IOBase)
        if mode is None and hasattr(f, 'mode'):
            mode = f.mode
        self.mode = mode
        self.name = name
        self.closed = False
        super(RawWrapper, self).__init__()

    def __repr__(self):
        return "<IO wrapper for {0}>".format(self._f)

    def close(self):
        self._f.close()
        self.closed = True

    def fileno(self):
        return self._f.fileno()

    def flush(self):
        return self._f.flush()

    def isatty(self):
        return self._f.isatty()

    def seek(self, offset, whence=SEEK_SET):
        return self._f.seek(offset, whence)

    def readable(self):
        if hasattr(self._f, 'readable'):
            return self._f.readable()
        return 'r' in self.mode

    def writable(self):
        if hasattr(self._f, 'writeable'):
            return self._fs.writeable()
        return 'w' in self.mode

    def seekable(self):
        if hasattr(self._f, 'seekable'):
            return self._f.seekable()
        try:
            self.seek(0, SEEK_CUR)
        except IOError:
            return False
        else:
            return True

    def tell(self):
        return self._f.tell()

    def truncate(self, size=None):
        return self._f.truncate(size)

    def write(self, data):
        if self.is_io:
            return self._f.write(data)
        self._f.write(data)
        return len(data)

    def read(self, n=-1):
        if n == -1:
            return self.readall()
        return self._f.read(n)

    def read1(self, n=-1):
        if self.is_io:
            return self._f.read1(n)
        return self.read(n)

    def readall(self):
        return self._f.read()

    def readinto(self, b):
        if self.is_io:
            return self._f.readinto(b)
        data = self._f.read(len(b))
        bytes_read = len(data)
        b[:len(data)] = data
        return bytes_read

    def readline(self, limit=-1):
        return self._f.readline(limit)

    def readlines(self, hint=-1):
        return self._f.readlines(hint)

    def writelines(self, sequence):
        return self._f.writelines(sequence)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def __iter__(self):
        return iter(self._f)


def filelike_to_stream(f):
    @wraps(f)
    def wrapper(self, path, mode='rt', buffering=-1, encoding=None, errors=None, newline=None, line_buffering=False, **kwargs):
        file_like = f(self,
                      path,
                      mode=mode,
                      buffering=buffering,
                      encoding=encoding,
                      errors=errors,
                      newline=newline,
                      line_buffering=line_buffering,
                      **kwargs)
        return make_stream(path,
                           file_like,
                           mode=mode,
                           buffering=buffering,
                           encoding=encoding,
                           errors=errors,
                           newline=newline,
                           line_buffering=line_buffering)
    return wrapper


def make_stream(name,
                f,
                mode='r',
                buffering=-1,
                encoding=None,
                errors=None,
                newline=None,
                line_buffering=False,
                **kwargs):
    """Take a Python 2.x binary file and returns an IO Stream"""
    r, w, a, binary = 'r' in mode, 'w' in mode, 'a' in mode, 'b' in mode
    if '+' in mode:
        r, w = True, True

    io_object = RawWrapper(f, mode=mode, name=name)
    if buffering >= 0:
        if r and w:
            io_object = io.BufferedRandom(io_object, buffering or io.DEFAULT_BUFFER_SIZE)
        elif r:
            io_object = io.BufferedReader(io_object, buffering or io.DEFAULT_BUFFER_SIZE)
        elif w:
            io_object = io.BufferedWriter(io_object, buffering or io.DEFAULT_BUFFER_SIZE)

    if not binary:
        io_object = io.TextIOWrapper(io_object,
                                     encoding=encoding or 'utf-8',
                                     errors=errors,
                                     newline=newline,
                                     line_buffering=line_buffering,)

    return io_object


def decode_binary(data, encoding=None, errors=None, newline=None):
    """Decode bytes as though read from a text file"""
    return io.TextIOWrapper(io.BytesIO(data), encoding=encoding or 'utf-8', errors=errors, newline=newline).read()


def make_bytes_io(data, encoding=None, errors=None):
    """Make a bytes IO object from either a string or an open file"""
    if hasattr(data, 'mode') and 'b' in data.mode:
        # It's already a binary file
        return data
    if not isinstance(data, basestring):
        # It's a file, but we don't know if its binary
        # TODO: Is there a better way than reading the entire file?
        data = data.read() or b''
    if isinstance(data, six.text_type):
        # If its text, encoding in to bytes
        data = data.encode(encoding=encoding, errors=errors)
    return io.BytesIO(data)


def copy_file_to_fs(f, fs, path, encoding=None, errors=None, progress_callback=None, chunk_size=64 * 1024):
    """Copy an open file to a path on an FS"""
    if progress_callback is None:
        progress_callback = lambda bytes_written: None
    read = f.read
    chunk = read(chunk_size)
    if isinstance(chunk, six.text_type):
        f = fs.open(path, 'wt', encoding=encoding, errors=errors)
    else:
        f = fs.open(path, 'wb')
    write = f.write
    bytes_written = 0
    try:
        while chunk:
            write(chunk)
            bytes_written += len(chunk)
            progress_callback(bytes_written)
            chunk = read(chunk_size)
    finally:
        f.close()
    return bytes_written


def line_iterator(f, size=None):
    """A not terribly efficient char by char line iterator"""
    read = f.read
    line = []
    append = line.append
    c = 1
    if size is None or size < 0:
        while c:
            c = read(1)
            if c:
                append(c)
            if c in (b'\n', b''):
                yield b''.join(line)
                del line[:]
    else:
        while c:
            c = read(1)
            if c:
                append(c)
            if c in (b'\n', b'') or len(line) >= size:
                yield b''.join(line)
                del line[:]


if __name__ == "__main__":
    print("Reading a binary file")
    bin_file = open('tests/data/UTF-8-demo.txt', 'rb')
    with make_stream('UTF-8-demo.txt', bin_file, 'rb') as f:
        print(repr(f))
        print(type(f.read(200)))

    print("Reading a text file")
    bin_file = open('tests/data/UTF-8-demo.txt', 'rb')
    with make_stream('UTF-8-demo.txt', bin_file, 'rt') as f:
        print(repr(f))
        print(type(f.read(200)))

    print("Reading a buffered binary file")
    bin_file = open('tests/data/UTF-8-demo.txt', 'rb')
    with make_stream('UTF-8-demo.txt', bin_file, 'rb', buffering=0) as f:
        print(repr(f))
        print(type(f.read(200)))



""" The model for list boxes. """

# Enthought library imports
from traits.api import Event, HasTraits


# Classes for event traits.
class ListModelEvent(object):
    """ Information about list model changes. """


class ListBoxModel(HasTraits):
    """ The model for list boxes. """

    #### Events ####

    #: Fired when the contents of the list have changed.
    list_changed = Event

    def get_item_count(self):
        """ Get the number of items in the list.

        Returns
        -------
        item_count : int
            The number of items in the list.
        """
        raise NotImplementedError

    def get_item_at(self, index):
        """ Returns the item at the specified index.

        Parameters
        ----------
        index : int
            The index to return the value of.

        Returns
        -------
        label, item : str, any
            The user-visible string and model data of the item.
        """
        raise NotImplementedError

    def fire_list_changed(self):
        """ Invoke this method when the list has changed. """
        self.list_changed = ListModelEvent()



# =============================================================================
# OWSLib. Copyright (C) 2005 Sean C. Gillies
#
# Contact email: sgillies@frii.com
# =============================================================================

from __future__ import (absolute_import, division, print_function)
import six
import inspect

def patch_well_known_namespaces(etree_module):

    import warnings
    from owslib.namespaces import Namespaces
    ns = Namespaces()

    """Monkey patches the etree module to add some well-known namespaces."""

    try:
        register_namespace = etree_module.register_namespace
    except AttributeError:
        try:
            etree_module._namespace_map

            def register_namespace(prefix, uri):
                etree_module._namespace_map[uri] = prefix
        except AttributeError:
            def register_namespace(prefix, uri):
                pass
            warnings.warn("Only 'lxml.etree' >= 2.3 and 'xml.etree.ElementTree' >= 1.3 are fully supported!")

    for k, v in six.iteritems(ns.get_namespaces()):
        register_namespace(k, v)

# try to find lxml or elementtree
try:
    from lxml import etree
    from lxml.etree import ParseError
    ElementType = etree._Element
except ImportError:
    try:
        # Python 2.x/3.x with ElementTree included
        import xml.etree.ElementTree as etree

        try:
            from xml.etree.ElementTree import ParseError
        except ImportError:
            from xml.parsers.expat import ExpatError as ParseError

        if hasattr(etree, 'Element') and inspect.isclass(etree.Element):
            # python 3.4, 3.3, 2.7
            ElementType = etree.Element
        else:
            # python 2.6
            ElementType = etree._ElementInterface

    except ImportError:
        try:
            # Python < 2.5 with ElementTree installed
            import elementtree.ElementTree as etree
            ParseError = StandardError      # i can't find a ParseError related item in elementtree docs!
            ElementType = etree.Element
        except ImportError:
            raise RuntimeError('You need either lxml or ElementTree to use OWSLib!')

patch_well_known_namespaces(etree)



class ProviderNotRegisteredError(Exception):
    """ Raised when a provider is requested by ID and not found. """

    def __init__(self, provider_id):
        self.provider_id = provider_id

    def __str__(self):
        msg = "Selection provider with ID '{id}' not found."
        return msg.format(id=self.provider_id)


class IDConflictError(Exception):
    """ Raised when a provider is added and its ID is already registered. """

    def __init__(self, provider_id):
        self.provider_id = provider_id

    def __str__(self):
        msg = "A selection provider with ID '{id}' is already registered."
        return msg.format(id=self.provider_id)


class ListenerNotConnectedError(Exception):
    """ Raised when a listener that was never connected is disconnected. """

    def __init__(self, provider_id, listener):
        self.provider_id = provider_id
        self.listener = listener

    def __str__(self):
        msg = "Selection listener {l} is not connected to provider '{id}'."
        return msg.format(l=self.listener, id=self.provider_id)



"""This module provides code that allows one to pickle the state of a
Python object to a dictionary.

The motivation for this is simple.  The standard Python
pickler/unpickler is best used to pickle simple objects and does not
work too well for complex code.  Specifically, there are two major
problems (1) the pickle file format is not easy to edit with a text
editor and (2) when a pickle is unpickled, it creates all the
necessary objects and sets the state of these objects.

Issue (2) might not appear to be a problem.  However, often, the
determination of the entire 'state' of an application requires the
knowledge of the state of many objects that are not really in the
users concern.  The user would ideally like to pickle just what he
thinks is relevant.  Now, given that the user is not going to save the
entire state of the application, the use of pickle is insufficient
since the state is no longer completely known (or worth knowing).  The
default `Unpickler` recreates the objects and the typical
implementation of `__setstate__` is usually to simply update the
object's `__dict__` attribute.  This is inadequate because the pickled
information is taken out of the real context when it was saved.

The `StatePickler` basically pickles the 'state' of an object into a
large dictionary.  This pickled data may be easily unpickled and
modified on the interpreter or edited with a text editor
(`pprint.saferepr` is a friend).  The second problem is also
eliminated.  When this state is unpickled using `StateUnpickler`, what
you get is a special dictionary (a `State` instance).  This allows one
to navigate the state just like the original object.  Its up to the
user to create any new objects and set their states using this
information.  This allows for a lot of flexibility while allowing one
to save and set the state of (almost) any Python object.

The `StateSetter` class helps set the state of a known instance.  When
setting the state of an instance it checks to see if there is a
`__set_pure_state__` method that in turn calls `StateSetter.set`
appropriately.

Additionally, there is support for versioning.  The class' version is
obtain from the `__version__` class attribute.  This version along
with the versions of the bases of a class is embedded into the
metadata of the state and stored.  By using `version_registry.py` a
user may register a handler for a particular class and module.  When
the state of an object is set using `StateSetter.set_state`, then
these handlers are called in reverse order of their MRO.  This gives
the handler an opportunity to upgrade the state depending on its
version.  Builtin classes are not scanned for versions.  If a class
has no version, then by default it is assumed to be -1.


Example::

  >>> class A:
  ...    def __init__(self):
  ...        self.a = 'a'
  ...
  >>> a = A()
  >>> a.a = 100
  >>> import state_pickler
  >>> s = state_pickler.dumps(a)               # Dump the state of `a`.
  >>> state = state_pickler.loads_state(s)     # Get the state back.
  >>> b = state_pickler.create_instance(state) # Create the object.
  >>> state_pickler.set_state(b, state)        # Set the object's state.
  >>> assert b.a == 100


Features
--------

 - The output is a plain old dictionary so is easy to parse, edit etc.
 - Handles references to avoid duplication.
 - Gzips Numeric arrays when dumping them.
 - Support for versioning.


Caveats
-------

 - Does not pickle a whole bunch of stuff including code objects and
   functions.
 - The output is a pure dictionary and does not contain instances.  So
   using this *as it is* in `__setstate__` will not work.  Instead
   define a `__set_pure_state__` and use the `StateSetter` class or
   the `set_state` function provided by this module.


Notes
-----

  Browsing the code from XMarshaL_ and pickle.py proved useful for
  ideas.  None of the code is taken from there though.

.. _XMarshaL:  http://www.dezentral.de/soft/XMarshaL

"""
# Author: Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) 2005-2015, Enthought, Inc.
# License: BSD Style.

# Standard library imports.
import base64
import sys
import types
import pickle
import gzip
from io import BytesIO, StringIO

import numpy

# Local imports.
from . import version_registry
from .file_path import FilePath

PY_VER = sys.version_info[0]
NumpyArrayType = type(numpy.array([]))


def gzip_string(data):
    """Given a string (`data`) this gzips the string and returns it.
    """
    s = BytesIO()
    writer = gzip.GzipFile(mode='wb', fileobj=s)
    writer.write(data)
    writer.close()
    s.seek(0)
    return s.read()


def gunzip_string(data):
    """Given a gzipped string (`data`) this unzips the string and
    returns it.
    """
    if PY_VER== 2 or (bytes is not str and type(data) is bytes):
        s = BytesIO(data)
    else:
        s = StringIO(data)
    writer = gzip.GzipFile(mode='rb', fileobj=s)
    data = writer.read()
    writer.close()
    return data

class StatePicklerError(Exception):
    pass

class StateUnpicklerError(Exception):
    pass

class StateSetterError(Exception):
    pass

######################################################################
# `State` class
######################################################################
class State(dict):
    """Used to encapsulate the state of an instance in a very
    convenient form.  The '__metadata__' attribute/key is a dictionary
    that has class specific details like the class name, module name
    etc.
    """
    def __init__(self, **kw):
        dict.__init__(self, **kw)
        self.__dict__ = self

######################################################################
# `StateDict` class
######################################################################
class StateDict(dict):
    """Used to encapsulate a dictionary stored in a `State` instance.
    The has_instance attribute specifies if the dict has an instance
    embedded in it.
    """
    def __init__(self, **kw):
        dict.__init__(self, **kw)
        self.has_instance = False

######################################################################
# `StateList` class
######################################################################
class StateList(list):
    """Used to encapsulate a list stored in a `State` instance.  The
    has_instance attribute specifies if the list has an instance
    embedded in it.
    """
    def __init__(self, seq=None):
        if seq:
            list.__init__(self, seq)
        else:
            list.__init__(self)
        self.has_instance = False

######################################################################
# `StateTuple` class
######################################################################
class StateTuple(tuple):
    """Used to encapsulate a tuple stored in a `State` instance.  The
    has_instance attribute specifies if the tuple has an instance
    embedded in it.
    """
    def __new__(cls, seq=None):
        if seq:
            obj = super(StateTuple, cls).__new__(cls, tuple(seq))
        else:
            obj = super(StateTuple, cls).__new__(cls)
        obj.has_instance = False
        return obj


######################################################################
# `StatePickler` class
######################################################################
class StatePickler:
    """Pickles the state of an object into a dictionary.  The
    dictionary is itself either saved as a pickled file (`dump`) or
    pickled string (`dumps`).  Alternatively, the `dump_state` method
    will return the dictionary that is pickled.

    The format of the state dict is quite strightfoward.  Basic types
    (bool, int, long, float, complex, None, string and unicode) are
    represented as they are.  Everything else is stored as a
    dictionary containing metadata information on the object's type
    etc. and also the actual object in the 'data' key.  For example::

        >>> p = StatePickler()
        >>> p.dump_state(1)
        1
        >>> l = [1,2.0, None, [1,2,3]]
        >>> p.dump_state(l)
        {'data': [1, 2.0, None, {'data': [1, 2, 3], 'type': 'list', 'id': 1}],
         'id': 0,
         'type': 'list'}

    Classes are also represented similarly.  The state in this case is
    obtained from the `__getstate__` method or from the `__dict__`.
    Here is an example::

        >>> class A:
        ...     __version__ = 1  # State version
        ...     def __init__(self):
        ...         self.attribute = 1
        ...
        >>> a = A()
        >>> p = StatePickler()
        >>> p.dump_state(a)
        {'class_name': 'A',
         'data': {'data': {'attribute': 1}, 'type': 'dict', 'id': 2},
         'id': 0,
         'initargs': {'data': (), 'type': 'tuple', 'id': 1},
         'module': '__main__',
         'type': 'instance',
         'version': [(('A', '__main__'), 1)]}

    When pickling data, references are taken care of.  Numeric arrays
    can be pickled and are stored as a gzipped base64 encoded string.

    """
    def __init__(self):
        self._clear()
        type_map = {bool: self._do_basic_type,
                    complex: self._do_basic_type,
                    float: self._do_basic_type,
                    int: self._do_basic_type,
                    type(None): self._do_basic_type,
                    str: self._do_basic_type,
                    bytes: self._do_basic_type,
                    tuple: self._do_tuple,
                    list: self._do_list,
                    dict: self._do_dict,
                    NumpyArrayType: self._do_numeric,
                    State: self._do_state,
                    }
        if PY_VER == 2:
            type_map[long] = self._do_basic_type
            type_map[unicode] = self._do_basic_type
        self.type_map = type_map

    def dump(self, value, file):
        """Pickles the state of the object (`value`) into the passed
        file.
        """
        try:
            # Store the file name we are writing to so we can munge
            # file paths suitably.
            self.file_name = file.name
        except AttributeError:
            pass
        pickle.dump(self._do(value), file)

    def dumps(self, value):
        """Pickles the state of the object (`value`) and returns a
        string.
        """
        return pickle.dumps(self._do(value))

    def dump_state(self, value):
        """Returns a dictionary or a basic type representing the
        complete state of the object (`value`).

        This value is pickled by the `dump` and `dumps` methods.
        """
        return self._do(value)

    ######################################################################
    # Non-public methods
    ######################################################################
    def _clear(self):
        # Stores the file name of the file being used to dump the
        # state.  This is used to change any embedded paths relative
        # to the saved file.
        self.file_name = ''
        # Caches id's to handle references.
        self.obj_cache = {}
        # Misc cache to cache things that are not persistent.  For
        # example, object.__getstate__()/__getinitargs__() usually
        # returns a copy of a dict/tuple that could possibly be reused
        # on another object's __getstate__.  Caching these prevents
        # some wierd problems with the `id` of the object.
        self._misc_cache = []

    def _flush_traits(self, obj):
        """Checks if the object has traits and ensures that the traits
        are set in the `__dict__` so we can pickle it.
        """
        # Not needed with Traits3.
        return

    def _do(self, obj):
        obj_type = type(obj)
        key = self._get_id(obj)
        if key in self.obj_cache:
            return self._do_reference(obj)
        elif obj_type in self.type_map:
            return self.type_map[obj_type](obj)
        elif isinstance(obj, tuple):
            # Takes care of StateTuples.
            return self._do_tuple(obj)
        elif isinstance(obj, list):
            # Takes care of TraitListObjects.
            return self._do_list(obj)
        elif isinstance(obj, dict):
            # Takes care of TraitDictObjects.
            return self._do_dict(obj)
        elif hasattr(obj, '__dict__'):
            return self._do_instance(obj)

    def _get_id(self, value):
        try:
            key = hash(value)
        except TypeError:
            key = id(value)
        return key

    def _register(self, value):
        key = self._get_id(value)
        cache = self.obj_cache
        idx = len(cache)
        cache[key] = idx
        return idx

    def _do_basic_type(self, value):
        return value

    def _do_reference(self, value):
        key = self._get_id(value)
        idx = self.obj_cache[key]
        return dict(type='reference', id=idx, data=None)

    def _do_instance(self, value):
        # Flush out the traits.
        self._flush_traits(value)

        # Setup the relative paths of FilePaths before dumping.
        if self.file_name and isinstance(value, FilePath):
            value.set_relative(self.file_name)

        # Get the initargs.
        args = ()
        if hasattr(value, '__getinitargs__') and value.__getinitargs__:
            args = value.__getinitargs__()

        # Get the object state.
        if hasattr(value, '__get_pure_state__'):
            state = value.__get_pure_state__()
        elif hasattr(value, '__getstate__'):
            state = value.__getstate__()
        else:
            state = value.__dict__

        state.pop('__traits_version__', None)

        # Cache the args and state since they are likely to be gc'd.
        self._misc_cache.extend([args, state])
        # Register and process.
        idx = self._register(value)
        args_data = self._do(args)
        data = self._do(state)

        # Get the version of the object.
        version = version_registry.get_version(value)
        module = value.__class__.__module__
        class_name = value.__class__.__name__

        return dict(type='instance',
                    module=module,
                    class_name=class_name,
                    version=version,
                    id=idx,
                    initargs=args_data,
                    data=data)

    def _do_state(self, value):
        metadata = value.__metadata__
        args = metadata.get('initargs')
        state = dict(value)
        state.pop('__metadata__')

        self._misc_cache.extend([args, state])

        idx = self._register(value)
        args_data = self._do(args)
        data = self._do(state)

        return dict(type='instance',
                    module=metadata['module'],
                    class_name=metadata['class_name'],
                    version=metadata['version'],
                    id=idx,
                    initargs=args_data,
                    data=data)

    def _do_tuple(self, value):
        idx = self._register(value)
        data = tuple([self._do(x) for x in value])
        return dict(type='tuple', id=idx, data=data)

    def _do_list(self, value):
        idx = self._register(value)
        data = [self._do(x) for x in value]
        return dict(type='list', id=idx, data=data)

    def _do_dict(self, value):
        idx = self._register(value)
        vals = [self._do(x) for x in value.values()]
        data = dict(zip(value.keys(), vals))
        return dict(type='dict', id=idx, data=data)

    def _do_numeric(self, value):
        idx = self._register(value)
        if PY_VER > 2:
            data = base64.encodebytes(gzip_string(numpy.ndarray.dumps(value)))
        else:
            data = base64.encodestring(gzip_string(numpy.ndarray.dumps(value)))
        return dict(type='numeric', id=idx, data=data)



######################################################################
# `StateUnpickler` class
######################################################################
class StateUnpickler:
    """Unpickles the state of an object saved using StatePickler.

    Please note that unlike the standard Unpickler, no instances of
    any user class are created.  The data for the state is obtained
    from the file or string, reference objects are setup to refer to
    the same state value and this state is returned in the form
    usually in the form of a dictionary.  For example::

        >>> class A:
        ...     def __init__(self):
        ...         self.attribute = 1
        ...
        >>> a = A()
        >>> p = StatePickler()
        >>> s = p.dumps(a)
        >>> up = StateUnpickler()
        >>> state = up.loads_state(s)
        >>> state.__class__.__name__
        'State'
        >>> state.attribute
        1
        >>> state.__metadata__
        {'class_name': 'A',
         'has_instance': True,
         'id': 0,
         'initargs': (),
         'module': '__main__',
         'type': 'instance',
         'version': [(('A', '__main__'), -1)]}

    Note that the state is actually a `State` instance and is
    navigable just like the original object.  The details of the
    instance are stored in the `__metadata__` attribute.  This is
    highly convenient since it is possible for someone to view and
    modify the state very easily.
    """

    def __init__(self):
        self._clear()
        self.type_map = {'reference': self._do_reference,
                         'instance': self._do_instance,
                         'tuple': self._do_tuple,
                         'list': self._do_list,
                         'dict': self._do_dict,
                         'numeric': self._do_numeric,
                         }

    def load_state(self, file):
        """Returns the state of an object loaded from the pickled data
        in the given file.
        """
        try:
            self.file_name = file.name
        except AttributeError:
            pass
        data = pickle.load(file)
        result = self._process(data)
        return result

    def loads_state(self, string):
        """Returns the state of an object loaded from the pickled data
        in the given string.
        """
        data = pickle.loads(string)
        result = self._process(data)
        return result

    ######################################################################
    # Non-public methods
    ######################################################################
    def _clear(self):
        # The file from which we are being loaded.
        self.file_name = ''
        # Cache of the objects.
        self._obj_cache = {}
        # Paths to the instances.
        self._instances = []
        # Caches the references.
        self._refs = {}
        # Numeric arrays.
        self._numeric = {}

    def _set_has_instance(self, obj, value):
        if isinstance(obj, State):
            obj.__metadata__['has_instance'] = value
        elif isinstance(obj, (StateDict, StateList, StateTuple)):
            obj.has_instance = value

    def _process(self, data):
        result = self._do(data)

        # Setup all the Numeric arrays.  Do this first since
        # references use this.
        for key, (path, val) in self._numeric.items():
            if isinstance(result, StateTuple):
                result = list(result)
                exec('result%s = val'%path)
                result = StateTuple(result)
            else:
                exec('result%s = val'%path)

        # Setup the references so they really are references.
        for key, paths in self._refs.items():
            for path in paths:
                x = self._obj_cache[key]
                if isinstance(result, StateTuple):
                    result = list(result)
                    exec('result%s = x'%path)
                    result = StateTuple(result)
                else:
                    exec('result%s = x'%path)
                # if the reference is to an instance append its path.
                if isinstance(x, State):
                    self._instances.append(path)

        # Now setup the 'has_instance' attribute.  If 'has_instance'
        # is True then the object contains an instance somewhere
        # inside it.
        for path in self._instances:
            pth = path
            while pth:
                ns = {'result': result}
                exec('val = result%s'%pth, ns, ns)
                self._set_has_instance(ns['val'], True)
                end = pth.rfind('[')
                pth = pth[:end]
            # Now make sure that the first element also has_instance.
            self._set_has_instance(result, True)
        return result

    def _do(self, data, path=''):
        if type(data) is dict:
            return self.type_map[data['type']](data, path)
        else:
            return data

    def _do_reference(self, value, path):
        id = value['id']
        if id in self._refs:
            self._refs[id].append(path)
        else:
            self._refs[id] = [path]
        return State(__metadata__=value)

    def _handle_file_path(self, value):
        if (value['class_name'] == 'FilePath') and \
           ('file_path' in value['module']) and \
           self.file_name:
            data = value['data']['data']
            fp = FilePath(data['rel_pth'])
            fp.set_absolute(self.file_name)
            data['abs_pth'] = fp.abs_pth

    def _do_instance(self, value, path):
        self._instances.append(path)
        initargs = self._do(value['initargs'],
                            path + '.__metadata__["initargs"]')
        # Handle FilePaths.
        self._handle_file_path(value)

        d = self._do(value['data'], path)
        md = dict(type='instance',
                  module=value['module'],
                  class_name=value['class_name'],
                  version=value['version'],
                  id=value['id'],
                  initargs=initargs,
                  has_instance=True)
        result = State(**d)
        result.__metadata__ = md
        self._obj_cache[value['id']] = result
        return result

    def _do_tuple(self, value, path):
        res = []
        for i, x in enumerate(value['data']):
            res.append(self._do(x, path + '[%d]'%i))
        result = StateTuple(res)
        self._obj_cache[value['id']] = result
        return result

    def _do_list(self, value, path):
        result = StateList()
        for i, x in enumerate(value['data']):
            result.append(self._do(x, path + '[%d]'%i))
        self._obj_cache[value['id']] = result
        return result

    def _do_dict(self, value, path):
        result = StateDict()
        for key, val in value['data'].items():
            result[key] = self._do(val, path + '["%s"]'%key)
        self._obj_cache[value['id']] = result
        return result

    def _do_numeric(self, value, path):
        if PY_VER > 2:
            data = value['data']
            if isinstance(data, str):
                data = value['data'].encode('utf-8')
            junk = gunzip_string(base64.decodebytes(data))
            result = numpy.loads(junk, encoding='bytes')
        else:
            junk = gunzip_string(value['data'].decode('base64'))
            result = numpy.loads(junk)
        self._numeric[value['id']] = (path, result)
        self._obj_cache[value['id']] = result
        return result


######################################################################
# `StateSetter` class
######################################################################
class StateSetter:
    """This is a convenience class that helps a user set the
    attributes of an object given its saved state.  For instances it
    checks to see if a `__set_pure_state__` method exists and calls
    that when it sets the state.
    """
    def __init__(self):
        # Stores the ids of instances already done.
        self._instance_ids = []
        self.type_map = {State: self._do_instance,
                         StateTuple: self._do_tuple,
                         StateList: self._do_list,
                         StateDict: self._do_dict,
                         }

    def set(self, obj, state, ignore=None, first=None, last=None):
        """Sets the state of the object.

        This is to be used as a means to simplify loading the state of
        an object from its `__setstate__` method using the dictionary
        describing its state.  Note that before the state is set, the
        registered handlers for the particular class are called in
        order to upgrade the version of the state to the latest
        version.

        Parameters
        ----------

        - obj : `object`

          The object whose state is to be set.  If this is `None`
          (default) then the object is created.

        - state : `dict`

          The dictionary representing the state of the object.

        - ignore : `list(str)`

          The list of attributes specified in this list are ignored
          and the state of these attributes are not set (this excludes
          the ones specified in `first` and `last`).  If one specifies
          a '*' then all attributes are ignored except the ones
          specified in `first` and `last`.

        - first : `list(str)`

          The list of attributes specified in this list are set first (in
          order), before any other attributes are set.

        - last : `list(str)`

          The list of attributes specified in this list are set last (in
          order), after all other attributes are set.

        """
        if (not isinstance(state, State)) and \
               state.__metadata__['type'] != 'instance':
            raise StateSetterError(
                'Can only set the attributes of an instance.'
            )

        # Upgrade the state to the latest using the registry.
        self._update_and_check_state(obj, state)

        self._register(obj)

        # This wierdness is needed since the state's own `keys` might
        # be set to something else.
        state_keys = list(dict.keys(state))
        state_keys.remove('__metadata__')

        if first is None:
            first = []
        if last is None:
            last = []

        # Remove all the ignored keys.
        if ignore:
            if '*' in ignore:
                state_keys = first + last
            else:
                for name in ignore:
                    try:
                        state_keys.remove(name)
                    except KeyError:
                        pass

        # Do the `first` attributes.
        for key in first:
            state_keys.remove(key)
            self._do(obj, key, state[key])

        # Remove the `last` attributes.
        for key in last:
            state_keys.remove(key)

        # Set the remaining attributes.
        for key in state_keys:
            self._do(obj, key, state[key])

        # Do the last ones in order.
        for key in last:
            self._do(obj, key, state[key])

    ######################################################################
    # Non-public methods.
    ######################################################################
    def _register(self, obj):
        idx = id(obj)
        if idx not in self._instance_ids:
            self._instance_ids.append(idx)

    def _is_registered(self, obj):
        return (id(obj) in self._instance_ids)

    def _has_instance(self, value):
        """Given something (`value`) that is part of the state this
        returns if the value has an instance embedded in it or not.
        """
        if isinstance(value, State):
            return True
        elif isinstance(value, (StateDict, StateList, StateTuple)):
            return value.has_instance
        return False

    def _get_pure(self, value):
        """Returns the Python representation of the object (usually a
        list, tuple or dict) that has no instances embedded within it.
        """
        result = value
        if self._has_instance(value):
            raise StateSetterError(
                'Value has an instance: %s'%value
            )
        if isinstance(value, (StateList, StateTuple)):
            result = [self._get_pure(x) for x in value]
            if isinstance(value, StateTuple):
                result = tuple(result)
        elif isinstance(value, StateDict):
            result = {}
            for k, v in value.items():
                result[k] = self._get_pure(v)
        return result

    def _update_and_check_state(self, obj, state):
        """Updates the state from the registry and then checks if the
        object and state have same class.
        """
        # Upgrade this state object to the latest using the registry.
        # This is done before testing because updating may change the
        # class name/module.
        version_registry.registry.update(state)

        # Make sure object and state have the same class and module names.
        metadata = state.__metadata__
        cls = obj.__class__
        if (metadata['class_name'] != cls.__name__):
            raise StateSetterError(
                'Instance (%s) and state (%s) do not have the same class'\
                ' name!'%(cls.__name__, metadata['class_name'])
            )
        if (metadata['module'] != cls.__module__):
            raise StateSetterError(
                'Instance (%s) and state (%s) do not have the same module'\
                ' name!'%(cls.__module__, metadata['module'])
            )

    def _do(self, obj, key, value):
        try:
            attr = getattr(obj, key)
        except AttributeError:
            raise StateSetterError(
                'Object %s does not have an attribute called: %s'%(obj, key)
            )

        if isinstance(value, (State, StateDict, StateList, StateTuple)):
            # Special handlers are needed.
            if not self._has_instance(value):
                result = self._get_pure(value)
                setattr(obj, key, result)
            elif isinstance(value, StateTuple):
                setattr(obj, key, self._do_tuple(getattr(obj, key), value))
            else:
                self._do_object(getattr(obj, key), value)
        else:
            setattr(obj, key, value)

    def _do_object(self, obj, state):
        self.type_map[state.__class__](obj, state)

    def _do_instance(self, obj, state):
        if self._is_registered(obj):
            return
        else:
            self._register(obj)

        metadata = state.__metadata__
        if hasattr(obj, '__set_pure_state__'):
            self._update_and_check_state(obj, state)
            obj.__set_pure_state__(state)
        elif 'tvtk_classes' in metadata['module']:
            self._update_and_check_state(obj, state)
            tmp = self._get_pure(StateDict(**state))
            del tmp['__metadata__']
            obj.__setstate__(tmp)
        else:
            # No need to update or check since `set` does it for us.
            self.set(obj, state)

    def _do_tuple(self, obj, state):
        if not self._has_instance(state):
            return self._get_pure(state)
        else:
            result = list(obj)
            self._do_list(result, state)
            return tuple(result)

    def _do_list(self, obj, state):
        if len(obj) == len(state):
            for i in range(len(obj)):
                if not self._has_instance(state[i]):
                    obj[i] = self._get_pure(state[i])
                elif isinstance(state[i], tuple):
                    obj[i] = self._do_tuple(state[i])
                else:
                    self._do_object(obj[i], state[i])
        else:
            raise StateSetterError(
                'Cannot set state of list of incorrect size.'
            )

    def _do_dict(self, obj, state):
        for key, value in state.items():
            if not self._has_instance(value):
                obj[key] = self._get_pure(value)
            elif isinstance(value, tuple):
                obj[key] = self._do_tuple(value)
            else:
                self._do_object(obj[key], value)


######################################################################
# Internal Utility functions.
######################################################################
def _get_file_read(f):
    if hasattr(f, 'read'):
        return f
    else:
        return open(f, 'rb')

def _get_file_write(f):
    if hasattr(f, 'write'):
        return f
    else:
        return open(f, 'wb')


######################################################################
# Utility functions.
######################################################################
def dump(value, file):
    """Pickles the state of the object (`value`) into the passed file
    (or file name).
    """
    f = _get_file_write(file)
    try:
        StatePickler().dump(value, f)
    finally:
        f.flush()
        if f is not file:
            f.close()


def dumps(value):
    """Pickles the state of the object (`value`) and returns a string.
    """
    return StatePickler().dumps(value)


def load_state(file):
    """Returns the state of an object loaded from the pickled data in
    the given file (or file name).
    """
    f = _get_file_read(file)
    try:
        state = StateUnpickler().load_state(f)
    finally:
        if f is not file:
            f.close()
    return state


def loads_state(string):
    """Returns the state of an object loaded from the pickled data
    in the given string.
    """
    return StateUnpickler().loads_state(string)


def get_state(obj):
    """Returns the state of the object (usually as a dictionary).  The
    returned state may be used directy to set the state of the object
    via `set_state`.
    """
    s = dumps(obj)
    return loads_state(s)


def set_state(obj, state, ignore=None, first=None, last=None):
    StateSetter().set(obj, state, ignore, first, last)
set_state.__doc__ = StateSetter.set.__doc__


def update_state(state):
    """Given the state of an object, this updates the state to the
    latest version using the handlers given in the version registry.
    The state is modified in-place.
    """
    version_registry.registry.update(state)


def create_instance(state):
    """Create an instance from the state if possible.
    """
    if (not isinstance(state, State)) and \
           ('class_name'  not in state.__metadata__):
        raise StateSetterError('No class information in state')
    metadata = state.__metadata__
    class_name = metadata.get('class_name')
    mod_name = metadata.get('module')
    if 'tvtk_classes' in mod_name:
        # FIXME: This sort of special-case is probably indicative of something
        # that needs more thought, plus it makes it tought to decide whether
        # this component depends on tvtk!
        from tvtk.api import tvtk
        return getattr(tvtk, class_name)()

    initargs = metadata['initargs']
    if initargs.has_instance:
        raise StateUnpicklerError('Cannot unpickle non-trivial initargs')

    __import__(mod_name, globals(), locals(), class_name)
    mod = sys.modules[mod_name]
    cls = getattr(mod, class_name)
    return cls(*initargs)



"""
fs.appdirfs
===========

A collection of filesystems that map to application specific locations.

These classes abstract away the different requirements for user data across platforms,
which vary in their conventions. They are all subclasses of :class:`fs.osfs.OSFS`,
all that differs from `OSFS` is the constructor which detects the appropriate
location given the name of the application, author name and other parameters.

Uses `appdirs` (https://github.com/ActiveState/appdirs), written by Trent Mick and Sridhar Ratnakumar <trentm at gmail com; github at srid name>

"""

from fs.osfs import OSFS
from fs.appdirs import AppDirs

__all__ = ['UserDataFS',
           'SiteDataFS',
           'UserCacheFS',
           'UserLogFS']


class UserDataFS(OSFS):
    """A filesystem for per-user application data."""
    def __init__(self, appname, appauthor=None, version=None, roaming=False, create=True):
        """
        :param appname: the name of the application
        :param appauthor: the name of the author (used on Windows)
        :param version: optional version string, if a unique location per version of the application is required
        :param roaming: if True, use a *roaming* profile on Windows, see http://technet.microsoft.com/en-us/library/cc766489(WS.10).aspx
        :param create: if True (the default) the directory will be created if it does not exist

        """
        app_dirs = AppDirs(appname, appauthor, version, roaming)
        super(UserDataFS, self).__init__(app_dirs.user_data_dir, create=create)


class SiteDataFS(OSFS):
    """A filesystem for application site data."""
    def __init__(self, appname, appauthor=None, version=None, roaming=False, create=True):
        """
        :param appname: the name of the application
        :param appauthor: the name of the author (not used on linux)
        :param version: optional version string, if a unique location per version of the application is required
        :param roaming: if True, use a *roaming* profile on Windows, see http://technet.microsoft.com/en-us/library/cc766489(WS.10).aspx
        :param create: if True (the default) the directory will be created if it does not exist

        """
        app_dirs = AppDirs(appname, appauthor, version, roaming)
        super(SiteDataFS, self).__init__(app_dirs.site_data_dir, create=create)


class UserCacheFS(OSFS):
    """A filesystem for per-user application cache data."""
    def __init__(self, appname, appauthor=None, version=None, roaming=False, create=True):
        """
        :param appname: the name of the application
        :param appauthor: the name of the author (not used on linux)
        :param version: optional version string, if a unique location per version of the application is required
        :param roaming: if True, use a *roaming* profile on Windows, see http://technet.microsoft.com/en-us/library/cc766489(WS.10).aspx
        :param create: if True (the default) the directory will be created if it does not exist

        """
        app_dirs = AppDirs(appname, appauthor, version, roaming)
        super(UserCacheFS, self).__init__(app_dirs.user_cache_dir, create=create)


class UserLogFS(OSFS):
    """A filesystem for per-user application log data."""
    def __init__(self, appname, appauthor=None, version=None, roaming=False, create=True):
        """
        :param appname: the name of the application
        :param appauthor: the name of the author (not used on linux)
        :param version: optional version string, if a unique location per version of the application is required
        :param roaming: if True, use a *roaming* profile on Windows, see http://technet.microsoft.com/en-us/library/cc766489(WS.10).aspx
        :param create: if True (the default) the directory will be created if it does not exist

        """
        app_dirs = AppDirs(appname, appauthor, version, roaming)
        super(UserLogFS, self).__init__(app_dirs.user_log_dir, create=create)


if __name__ == "__main__":
    udfs = UserDataFS('exampleapp', appauthor='pyfs')
    print udfs
    udfs2 = UserDataFS('exampleapp2', appauthor='pyfs', create=False)
    print udfs2





class SVGFilterInputs:
    SourceGraphic = 'SourceGraphic'
    SourceAlpha = 'SourceAlpha'
    BackgroundImage = 'BackgroundImage'
    BackgroundAlpha = 'BackgroundAlpha'
    FillPaint = 'FillPaint'
    StrokePaint = 'StrokePaint'


class SVGFilter:

    def __init__(self, svg):
        self.svg = svg

    def render(self, renderfn):
        # do some setup

        renderfn()

        # undo some setup


# -*- coding: ISO-8859-15 -*-
# =============================================================================
# Copyright (c) 2013 Tom Kralidis
#
# Authors : Tom Kralidis <tomkralidis@gmail.com>
#
# Contact email: tomkralidis@gmail.com
# =============================================================================

"""
Sensor Observation Service (SOS) methods and metadata. Factory function.
"""

from __future__ import (absolute_import, division, print_function)

from .swe.observation import sos100, sos200


def SensorObservationService(url, version='1.0.0', xml=None):
    """sos factory function, returns a version specific SensorObservationService object"""
    if version in  ['1.0', '1.0.0']:
        return sos100.SensorObservationService_1_0_0.__new__(sos100.SensorObservationService_1_0_0, url, version, xml)
    elif version in ['2.0', '2.0.0']:
        return sos200.SensorObservationService_2_0_0.__new__(sos200.SensorObservationService_2_0_0, url, version, xml)




from __future__ import absolute_import

from traits.testing.unittest_tools import unittest, UnittestTools

from ..widget import Widget


class TestWidget(unittest.TestCase, UnittestTools):

    def setUp(self):
        self.widget = Widget()

    def test_create(self):
        # create is not Implemented
        with self.assertRaises(NotImplementedError):
            self.widget._create()

    def test_destroy(self):
        # test that destroy works even when no control
        self.widget.destroy()



"""Sample panes for Skeleton

"""
# Enthought library imports.
from pyface.tasks.api import DockPane


class SkeletonPane1(DockPane):
    #### TaskPane interface ###################################################

    id = 'text_edit.pane1'
    name = 'Skeleton Pane 1'


class SkeletonPane2(DockPane):
    #### TaskPane interface ###################################################

    id = 'text_edit.pane2'
    name = 'Skeleton Pane 2'


class SkeletonPane3(DockPane):
    #### TaskPane interface ###################################################

    id = 'text_edit.pane3'
    name = 'Skeleton Pane 3'



"""
fs.osfs.xattrs
==============

Extended-attribute support for OSFS

"""


import os
import sys
import errno

from fs.errors import *
from fs.path import *
from fs.base import FS

try:
    import xattr
except ImportError:
    xattr = None


if xattr is not None:

    class OSFSXAttrMixin(object):
        """Mixin providing extended-attribute support via the 'xattr' module"""

        def __init__(self, *args, **kwargs):
            super(OSFSXAttrMixin, self).__init__(*args, **kwargs)

        @convert_os_errors
        def setxattr(self, path, key, value):
            xattr.xattr(self.getsyspath(path))[key]=value

        @convert_os_errors
        def getxattr(self, path, key, default=None):
            try:
                return xattr.xattr(self.getsyspath(path)).get(key)
            except KeyError:
                return default

        @convert_os_errors
        def delxattr(self, path, key):
            try:
                del xattr.xattr(self.getsyspath(path))[key]
            except KeyError:
                pass

        @convert_os_errors
        def listxattrs(self, path):
            return xattr.xattr(self.getsyspath(path)).keys()

else:

    class OSFSXAttrMixin(object):
        """Mixin disable extended-attribute support."""

        def __init__(self, *args, **kwargs):
            super(OSFSXAttrMixin, self).__init__(*args, **kwargs)

        def getxattr(self,path,key,default=None):
            raise UnsupportedError

        def setxattr(self,path,key,value):
            raise UnsupportedError

        def delxattr(self,path,key):
            raise UnsupportedError

        def listxattrs(self,path):
            raise UnsupportedError




#------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/07/2004
#
#------------------------------------------------------------------------------

""" Defines the text editor factory for all traits toolkit backends.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Dict, Str, Any, Bool

# CIRCULAR IMPORT FIXME: Importing from the source rather than traits.ui.api
# to avoid circular imports, as this EditorFactory will be part of
# traits.ui.api as well.
from ..view import View

from ..group import Group

from ..ui_traits import AView

from ..editor_factory import EditorFactory

#-------------------------------------------------------------------------
#  Define a simple identity mapping:
#-------------------------------------------------------------------------


class _Identity(object):
    """ A simple identity mapping.
    """

    def __call__(self, value):
        return value

#-------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------

# Mapping from user input text to other value
mapping_trait = Dict(Str, Any)

# Function used to evaluate textual user input
evaluate_trait = Any(_Identity())

#-------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------


class ToolkitEditorFactory(EditorFactory):
    """ Editor factory for text editors.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Dictionary that maps user input to other values
    mapping = mapping_trait

    # Is user input set on every keystroke?
    auto_set = Bool(True)

    # Is user input set when the Enter key is pressed?
    enter_set = Bool(False)

    # Is multi-line text allowed?
    multi_line = Bool(True)

    # Is editor readonly (will use custom / default editor appearance with readonly flag set to true)
    # in contrasrt with readonly style for item when completely another edito
    # is used
    read_only = Bool(False)

    # Is user input unreadable? (e.g., for a password)
    password = Bool(False)

    # Function to evaluate textual user input
    evaluate = evaluate_trait

    # The object trait containing the function used to evaluate user input
    evaluate_name = Str

    # The optional view to display when a read-only text editor is clicked:
    view = AView

    # In a read-only text editor, allow selection and copying of the text.
    readonly_allow_selection = Bool(False)

    #-------------------------------------------------------------------------
    #  Traits view definition:
    #-------------------------------------------------------------------------

    traits_view = View(['auto_set{Set value when text is typed}',
                        'enter_set{Set value when enter is pressed}',
                        'multi_line{Allow multiple lines of text}',
                        '<extras>',
                        '|options:[Options]>'])

    extras = Group('password{Is this a password field?}')

# Define the TextEditor class.
TextEditor = ToolkitEditorFactory

### EOF #######################################################################



#-----------------------------------------------------------------------------
# Name:        springtabs.py
# Purpose:     Tab-bar control that pops up windows when clicked
#
# Author:      Rob McMullen
#
# Created:     2008
# RCS-ID:      $Id: $
# Copyright:   (c) 2007 Rob McMullen
# License:     wxWidgets
#-----------------------------------------------------------------------------
"""SpringTabs

This module provides popup windows from a group of tabs.

Sometimes the popup won't properly close when clicking outside of the popup.
Certain controls seem to swallow the EVT_KILL_FOCUS event and without that
event propagating upwards, the sidebar handler never closes an open sidebar.
This seems to happen when the popup includes buttons or controls that are a
descendant of wx.PyControl.

In these cases, if the popup class defines the instance attribute
"lose_focus_helper_function" and binds all the offending controls to the
wx.EVT_KILL_FOCUS event with a callback function that calls the helper
function, the helper function will cause the popup to be dismissed.

class MyCoolSidebar(wx.ScrolledWindow):
    def __init__(...)
        ...
        btn = wx.Button(self, -i, "Button")
        btn.Bind(wx.EVT_KILL_FOCUS, self.on_lose_child_focus)

    def on_lose_child_focus(self, evt):
        evt.Skip()
        if self.lose_focus_helper_function is not None:
            self.lose_focus_helper_function(evt)

"""

import os, sys, struct, Queue, threading, time, socket
from cStringIO import StringIO

import wx
import wx.stc
from wx.lib.buttons import GenToggleButton

import logging
log = logging.getLogger(__name__)


class FakePopupWindow(wx.MiniFrame):
    def __init__(self, parent, style=None):
        super(FakePopupWindow, self).__init__(parent, style = wx.NO_BORDER |wx.FRAME_FLOAT_ON_PARENT | wx.FRAME_NO_TASKBAR)
        #self.Bind(wx.EVT_KEY_DOWN , self.OnKeyDown)
        self.Bind(wx.EVT_CHAR, self.on_char)
        self.Bind(wx.EVT_SET_FOCUS, self.on_focus)

    # Superclass overrides

    def Position(self, position, size):
        #print("pos=%s size=%s" % (position, size))
        self.Move((position[0]+size[0], position[1]+size[1]))

    def SetPosition(self, position):
        #print("pos=%s" % (position))
        self.Move((position[0], position[1]))

    # Compatibility methods for wx.PopupTransientWindow

    def Popup(self):
        self.Show(True)

    def Dismiss(self):
        self.Show(False)

    # local methods

    def fix_sizer(self, child):
        pass

    def on_char(self, evt):
        #print("on_char: keycode=%s" % evt.GetKeyCode())
        self.GetParent().GetEventHandler().ProcessEvent(evt)

    def activate_parent(self):
        """Activate the parent window
        @postcondition: parent window is raised

        """
        parent = self.GetParent()
        parent.Raise()
        parent.SetFocus()

    def on_focus(self, evt):
        """Raise and reset the focus to the parent window whenever
        we get focus.
        @param evt: event that called this handler

        """
        #log.debug("on_focus: set focus to %s" % str(self.GetParent()))
        self.activate_parent()
        evt.Skip()


class RealPopupWindowMac(wx.PopupTransientWindow):
    # Superclass overrides

    def OnDismiss(self):
        print "DISMISSED!!!!!"
        wx.CallAfter(self.GetParent().clear_popup)

    # local methods

    def fix_sizer(self, child):
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(child, 1, wx.EXPAND)
        self.SetSizer(sizer)


class RealPopupWindowWx(wx.PopupWindow):
    # Compatibility methods for wx.PopupTransientWindow

    def Popup(self):
        self.Show(True)

    def Dismiss(self):
        self.Show(False)

    # local methods

    def fix_sizer(self, child):
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(child, 1, wx.EXPAND)
        self.SetSizer(sizer)


if sys.platform == "darwin":
    RealPopupWindow = RealPopupWindowMac
else:
    RealPopupWindow = FakePopupWindow


class SpringTabItemRenderer(object):
    # local methods

    def __init__(self):
        self.spacing_between_items = 8

    def on_paint(self, item, evt):
        (width, height) = item.GetClientSizeTuple()
        x1 = y1 = 0
        x2 = width-1
        y2 = height-1

        dc = wx.PaintDC(item)
        if item.hover:
            self.draw_hover_background(item, dc)
        else:
            brush = item.GetBackgroundBrush(dc)
            if brush is not None:
                dc.SetBackground(brush)
                dc.Clear()

        item.draw_label(dc, width, height)
        self.draw_hover_decorations(item, dc, width, height)

        #log.debug("button %s: pressed=%s" % (item.GetLabel(), not item.up))

    def draw_hover_background(self, item, dc):
        brush = wx.Brush(item.faceDnClr, wx.SOLID)
        dc.SetBackground(brush)
        dc.Clear()

    def draw_hover_decorations(self, item, dc, width, height):
        pass

    def draw_notification(self, dc, x, y, w, h, num, springtabs):
        if num > 99:
            s = "99+"
        else:
            s = str(num)
        tw, th = dc.GetTextExtent(s)
        dc.SetBrush(springtabs.notification_brush)
        dc.SetPen(springtabs.notification_pen)
        dc.SetTextBackground(springtabs.notification_background)
        dc.SetTextForeground(springtabs.notification_text)
        if tw > w - h/2:
            w2 = tw + h/2
            x -= (w2 - w) / 2
            w = w2
        dc.DrawRoundedRectangle(x, y, w, h, h / 2)
        dc.DrawText(s, x + w/2 - tw/2, y + h/2 - th/2)


class SpringTabItemVerticalRenderer(SpringTabItemRenderer):
    default_direction = "right"

    def __init__(self, popup_direction="default"):
        SpringTabItemRenderer.__init__(self)
        if popup_direction == "default":
            popup_direction = self.default_direction

        if popup_direction == "left":
            self.popleft = True
        elif popup_direction == "right":
            self.popleft = False
        else:
            raise TypeError("popup_direction %s not valid for vertical renderer" % popup_direction)

    # local methods

    def get_best_size(self, item):
        """
        Overridden base class virtual.  Determines the best size of the
        button based on the label and bezel size.
        """
        h, w, useMin = item._GetLabelSize()
        width = w + 2 * item.border + item.bezelWidth - 1
        height = h + 2 * item.border + item.bezelWidth - 1 + self.spacing_between_items
        #log.debug("width=%d height=%d" % (width, height))
        return (width, height)

    def draw_label(self, item, dc, width, height, dx=0, dy=0):
        dc.SetFont(item.GetFont())
        if item.IsEnabled():
            dc.SetTextForeground(item.GetForegroundColour())
        else:
            dc.SetTextForeground(wx.SystemSettings.GetColour(wx.SYS_COLOUR_GRAYTEXT))
        label = item.GetLabel()
        th, tw = dc.GetTextExtent(label)
        #dc.DrawText(label, (width-tw)/2+dx, (height-th)/2+dy)
        if self.popleft:
            x = (width-tw)/2+dx + tw
            y = dy + item.border + self.spacing_between_items/2
            dc.DrawRotatedText(label, x, y, 270.0)
        else:
            x = (width-tw)/2+dx
            y = height + dy - item.border - self.spacing_between_items/2
            dc.DrawRotatedText(label, x, y, 90.0)
        num = item.notification_count
        if num > 0:
            parent = item.GetParent()
            dc.SetFont(parent.notification_font)
            tw, th = dc.GetTextExtent("0123456789")
            x = item.border
            w = width - 2*item.border
            h = th + 2*item.border
            if self.popleft:
                y = height - h - item.border
            else:
                y = item.border
            self.draw_notification(dc, x, y, w, h, num, parent)

    def set_popup_width(self, parent, popup, child):
        pw, ph = popup.GetSizeTuple()
        pcw, pch = popup.GetClientSizeTuple()
        try:
            cw, ch = child.DoGetBestSize()
        except AttributeError:
            cw, ch = child.GetSizeTuple()
        log.debug("popup size=%s  popup client size=%s  child=%s" % (str((pw, ph)), str((pcw, pch)), str((cw, ch))))

        # The client size may be smaller than the popup window if the popup
        # has a border decoration.
        diffwidth =  pw - pcw
        diffheight =  ph - pch

        # The popup will be at least as tall as the SpringTabs panel
        width, height = parent.GetSizeTuple()
        if ph < height:
            ph = height
        pw = min(cw + diffwidth, parent.max_popup_width)
        popup.SetSize(wx.Size(pw, ph))
        return width, height, pw, ph

    def show_popup(self, parent, item, show=True):
        popup, child = item.get_popup()
        if show:
            if hasattr(child, 'activate_spring_tab'):
                child.activate_spring_tab()

            # Calculate the child's width twice because the popup itself may
            # not have a valid height the first time through if the popup has
            # never been shown before, and the popup height will be needed by
            # the child to calculate if a scrollbar is needed.
            self.set_popup_width(parent, popup, child)
            width, height, pw, ph = self.set_popup_width(parent, popup, child)

            x, y = parent.ClientToScreenXY(width, 0)
            if self.popleft:
                x -= width + pw
            #log.debug("popping up at %s" % str((x,y)))
            child.SetPosition(wx.Point(0, 0))
            popup.SetPosition(wx.Point(x, y))
            wx.CallAfter(item.set_popup_lose_focus)
            popup.Popup()
        else:
            popup.Dismiss()


class SpringTabItem(GenToggleButton):
    def __init__(self, parent, id=-1, label='', window_cb=None, available_cb=None, **kwargs):
        """Creates a springtab button
        
        This button on the springtab is linked to a callback function that
        creates a window on demand when the button is clicked.
        
        @param parent: parent window: a L{SpringTab} instance
        
        @param id: optional id of the new button
        
        @param label: label of the springtab button
        
        @param window_cb: callback to create the contents of the popup.  This
        functor will be passed two arguments: the first is the popup window
        determined from L{SpringTab.get_new_popup} to be used as the parent
        window for the popup contents.  The second argument is the kwargs dict.
        
        @param available_cb: callback to see if the button should be displayed
        in the L{SpringTab}.  This callback takes a single argument; the
        kwargs dict and should return a boolean indicating whether or not the
        button will be visible.
        
        @param kwargs: dict containing any keyword arguments necessary to
        create the popup contents or check for button availability.
        """
        self.border = 2
        self.hover = False

        GenToggleButton.__init__(self, parent, id, label)
        self.Bind(wx.EVT_ENTER_WINDOW, self.on_enter)
        self.Bind(wx.EVT_LEAVE_WINDOW, self.on_leave)

        self.window_cb = window_cb
        self.available_cb = available_cb
        if self.available_cb:
            self.Show(self.available_cb(**kwargs))
        self.kwargs = kwargs
        self.popup = None
        self.notification_count = 0
        self.skip_next_lose_focus = False

    # Superclass overrides

    def InitColours(self):
        faceClr = self.GetBackgroundColour()
        r, g, b = faceClr.Get()
        fr, fg, fb = max(0,r-32), max(0,g-32), max(0,b-32)
        #log.debug(str((fr, fg, fb)))
        self.faceDnClr = wx.Colour(fr, fg, fb)
        sr, sg, sb = max(0,r-32), max(0,g-32), max(0,b-32)
        self.shadowPen = wx.Pen(wx.Colour(sr,sg,sb), 1, wx.SOLID)
        hr, hg, hb = min(255,r+64), min(255,g+64), min(255,b+64)
        self.highlightPen = wx.Pen(wx.Colour(hr,hg,hb), 1, wx.SOLID)
        self.focusClr = wx.Colour(hr, hg, hb)

    def DoGetBestSize(self):
        return self.GetParent().get_renderer().get_best_size(self)

    def OnPaint(self, evt):
        self.GetParent().get_renderer().on_paint(self, evt)

    def OnLeftDown(self, event):
        log.debug("clicked on %s" % self.GetLabel())
        if not self.IsEnabled():
            return
        self.saveUp = self.up
        self.up = not self.up
        self.GetParent().set_radio(self)
        self.SetFocus()
        self.Refresh()

    # local methods

    def draw_label(self, dc, width, height, dx=0, dy=0):
        self.GetParent().get_renderer().draw_label(self, dc, width, height, dx, dy)

    def set_toggle(self, flag, check_popup=True):
        self.up = not flag
        if self.up:
            # with PopupTransientWindow, need to force not hovering so it
            # doesn't take a second click to process
            self.hover = False
        if check_popup:
            self.GetParent().set_radio(self)
        self.Refresh()

    def on_enter(self, evt):
        self.hover = True
        self.Refresh()

    def on_leave(self, evt):
        self.hover = False
        self.Refresh()

    def get_popup(self):
        parent = self.GetParent()
        if self.popup is None:
            self.popup = parent.get_new_popup()
            self.popup.Bind(wx.EVT_ACTIVATE, self.on_activate)

            # Create the window using the popup as the parent
            self.window_cb(self.popup, parent.task, **self.kwargs)
            windowlist = self.popup.GetChildren()
            log.debug("Creating popup: %s" % windowlist)
            if len(windowlist) == 0:
                raise RuntimeError("Popup window creation failed!")
            child = windowlist[0]
            if hasattr(child, 'lose_focus_helper_function'):
                child.lose_focus_helper_function = self.on_lose_child_focus
            child.Bind(wx.EVT_KILL_FOCUS, self.on_lose_child_focus)
            if sys.platform != "darwin":
                child.Bind(wx.EVT_CHAR_HOOK, self.on_char_hook)
            if sys.platform == "linux2":
                child.Bind(wx.EVT_RIGHT_DOWN, self.on_right_down)
            self.popup.fix_sizer(child)
        windowlist = self.popup.GetChildren()
        if len(windowlist) == 0:
            raise RuntimeError("Popup window creation failed!")
        child = windowlist[0]
        return self.popup, child

    def on_char_hook(self, evt):
        """ESC handler that closes the popup

        """
        log.debug("on_char_hook: keycode=%s, popup=%s" % (evt.GetKeyCode(), evt.GetEventObject()))
        if evt.GetKeyCode() == wx.WXK_ESCAPE:
            wx.CallAfter(self.GetParent().cancel_popup)
        else:
            evt.Skip()

    def on_activate(self, evt):
        #log.debug("Activating %s: %s, shown=%s" % (self.GetLabel(), evt.GetActive(), self.popup.IsShown()))
        if self.popup.IsShown():
            evt.Skip()

    # Note: somehow, the name "set_popup_lose_focus_callback" was not working,
    # but removing the _callback worked. Something for another day.

    def set_popup_lose_focus(self):
        """Callback for use within wx.CallAfter to prevent focus being set
        after the control has been removed.
        """
        popup, child = self.get_popup()
        if popup.IsShown():
            #log.debug("setting focus to %s" % self.GetLabel())
            child.SetFocus()

    @property
    def is_shown(self):
        if self.popup is None:
            return False
        popup, child = self.get_popup()
        return popup.IsShown()

    @property
    def managed_window(self):
        if self.popup is None:
            return None
        popup, child = self.get_popup()
        return child

    def set_popup_lose_focus_callback(self):
        """Callback for use within wx.CallAfter to prevent focus being set
        after the control has been removed.
        """
        popup, child = self.get_popup()
        if popup.IsShown():
            log.debug("removing focus from %s" % self.GetLabel())
            self.GetParent().clear_popup()

    def on_lose_child_focus(self, evt):
        popup = evt.GetEventObject()
        focus = evt.GetWindow()
        log.debug("on_lose_child_focus: tab: %s, win=%s new=%s, top=%s" % (self.GetLabel(), popup, focus, wx.GetApp().GetTopWindow()))
        if self.skip_next_lose_focus:
            log.debug("Skipping next on_lose_child_focus; hack for popup menu!")
            self.skip_next_lose_focus = False
        elif popup is not None:
            if sys.platform == "linux2" or popup != focus:
                wx.CallAfter(self.set_popup_lose_focus_callback)

    def on_right_down(self, evt):
        self.skip_next_lose_focus = True
        evt.Skip()

    def recalc_notification(self):
        popup, child = self.get_popup()
        try:
            count = child.get_notification_count()
            if count != self.notification_count:
                self.notification_count = count
                self.Refresh()
        except AttributeError:
            pass


class SpringTabs(wx.Panel):
    max_popup_width = 800

    def __init__(self, parent, task, *args, **kwargs):
        # Need to remove foreign keyword arguments
        vertical = True
        if 'orientation' in kwargs:
            if kwargs['orientation'] == 'horizontal':
                vertical = False
            del kwargs['orientation']
        popup_direction = 'default'
        if 'popup_direction' in kwargs:
            popup_direction = kwargs['popup_direction']
            del kwargs['popup_direction']
        wx.Panel.__init__(self, parent, *args, **kwargs)
        self.task = task

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        self._tabs = []
        if vertical:
            self._tab_renderer = SpringTabItemVerticalRenderer(popup_direction)
        else:
            raise RuntimeError("horizontal tab rendering not supported yet.")
        self._radio = None
        self._debug_paint = False
        self._popup_cls = RealPopupWindow

        self.Bind(wx.EVT_PAINT, self.on_paint)
        self.Bind(wx.EVT_SIZE, self.on_size)

        self.init_colors()

    # local methods

    def on_size(self, evt):
        self.Refresh()
        evt.Skip()

    def init_colors(self):
        self.notification_background = wx.Colour(240, 120, 120)
        self.notification_brush = wx.Brush(self.notification_background, wx.SOLID)
        self.notification_pen = wx.Pen(self.notification_background, 1, wx.SOLID)
        self.notification_text = wx.Colour(255, 255, 255)
        self.notification_font = wx.Font(8, wx.FONTFAMILY_SWISS, wx.NORMAL, wx.NORMAL)

    def get_renderer(self):
        return self._tab_renderer

    def set_radio(self, item):
        self._processing_radio = True
        if self._radio != item:
            self.clear_popup()
            self.popup_item(item)
        elif not item.GetToggle():
            self.clear_popup()

    def get_new_popup(self):
        popup = self._popup_cls(self)
        return popup

    def popup_item(self, item):
        self._radio = item
        #log.debug("Popping up %s" % item.GetLabel())
        self._tab_renderer.show_popup(self, item)

    def clear_popup(self, data=None):
        # uncomment this to show where the clear_popup is being called
        #import traceback
        #log.debug("".join(traceback.format_stack()))
        if self._radio is not None:
            #log.debug("Removing popup %s" % self._radio.GetLabel())
            self._tab_renderer.show_popup(self, self._radio, False)
            self._radio.set_toggle(False, check_popup=False)
        self._radio = None
        self.Refresh()

    def has_popup(self):
        return self._radio is not None

    def on_paint(self, evt):
        if self._debug_paint:
            dc = wx.PaintDC(self)

            size = self.GetClientSize()
            dc.SetFont(wx.NORMAL_FONT)
            dc.SetBrush(wx.WHITE_BRUSH)
            dc.SetPen(wx.WHITE_PEN)
            dc.DrawRectangle(0, 0, size.x, size.y)
            dc.SetPen(wx.LIGHT_GREY_PEN)
            dc.DrawLine(0, 0, size.x, size.y)
            dc.DrawLine(0, size.y, size.x, 0)

        evt.Skip()

    def add_tab(self, title, window_create_callback, window_available_callback=None, **kwargs):
        tab = SpringTabItem(self, label=title, window_cb=window_create_callback,available_cb=window_available_callback, **kwargs)
        self.GetSizer().Add(tab, 0, wx.EXPAND)
        self._tabs.append(tab)
        self.Layout()
        self.Refresh()

    def has_tabs(self):
        return bool(self._tabs)

    def delete_tabs(self):
        self.clear_popup()
        for tab in self._tabs:
            tab.Destroy()
        self._tabs = []

    def update_notifications(self):
        for tab in self._tabs:
            tab.recalc_notification()

    def cancel_popup(self):
        self.task.on_hide_minibuffer_or_cancel(None)



if sys.platform == "darwin":
    StatusPopupWindow = RealPopupWindowMac
else:
    StatusPopupWindow = RealPopupWindowWx

class PopupStatusBar(StatusPopupWindow):
    """Transient status bar that displays status text in a popup
    
    Unlike a wx.StatusBar window that uses a constant amount of screen real-
    estate, the PopupStatusBar displays status info in a temporary popup
    that is displayed on the bottom of the frame.
    
    Status text is always overwritten with updates to the status text.  Status
    text is designed to be used for repetitive updates in response to an event;
    for example, to update coordinates when the mouse is moved.
    
    To display menu help text like the wx.StatusBar does by default, capture
    the wx.EVT_MENU_HIGHLIGHT event and send the help text to the status bar
    using a method like:
    
    def OnMenuHighlight(self, evt):
        menu_id = evt.GetMenuId()
        if menu_id >= 0:
            help_text = self.GetMenuBar().GetHelpString(menu_id)
            if help_text:
                self.popup_status.showStatusText(help_text)

    """
    def __init__(self, frame):
        """Creates (but doesn't show) the PopupStatusBar
        
        @param frame: the parent frame
        
        @kwarg delay: (optional) delay in milliseconds before each message
        decays
        """
        StatusPopupWindow.__init__(self, frame)
        self.SetBackgroundColour("#B6C1FF")

        self.stack = wx.BoxSizer(wx.VERTICAL)
        self.status = wx.StaticText(self, -1, "", style = wx.BORDER_NONE)
        self.stack.Add(self.status, 0, wx.ALL|wx.EXPAND, 0)
        self.SetSizer(self.stack)
        self.Hide()

    def show_status_text(self, text):
        """Display a status text string in the status popup.
        
        This method is intended to display text in response to a large number
        of updates to a similar actions, for example: updating x,y coordinates
        in response to mouse movement.  It is undesirable to keep these
        messages in the list as the list would quickly grow to display many
        lines.  Instead, status text updates replace any previous status
        updates at the bottom of the popup.
        
        This forces the popup to be displayed if it isn't currently displayed.
        If the popup is displayed and other messages are present, the existing
        messages are moved up and this status text is inserted at the bottom.
        
        @param text: message to display
        """
        text = text.replace("\r\n", " ").replace("\n", " ").replace("\r", " ")
        self.status.SetLabelText(text)
        if text:
            self.position_and_show()
        else:
            self.Hide()

    def position_and_show(self):
        #self.Hide()
        self.stack.Fit(self)
        #self.stack.SetSizeHints(self)
        self.Layout()
        frame = self.GetParent()
        frame_offset = frame.GetClientAreaOrigin()
        frame_pos = frame.ClientToScreenXY(frame_offset[0], frame_offset[1])
        frame_size = frame.GetClientSizeTuple()
        win_size = self.GetSizeTuple()
        #print("frame pos: %s, size=%s  popup size=%s" % (str(frame_pos), str(frame_size), str(win_size)))
        x = frame_pos[0]
        y = frame_pos[1] + frame_size[1] - win_size[1]
        self.Position((x, y), (0,0))
        if win_size[0] > frame_size[0]:
            cropped_size = (frame_size[0], win_size[1])
            #print("Cropped! %s" % str(cropped_size))
            self.SetSize(cropped_size)
        self.Show(True)

    def clear(self):
        """Remove all messages and hide the popup"""
        self.show_status_text("")


if __name__ == "__main__":
    import wx.calendar
    import wx.stc

    class FontList(wx.Panel):
        def __init__(self, parent, *args, **kwargs):
            wx.Panel.__init__(self, parent, -1)

            e = wx.FontEnumerator()
            e.EnumerateFacenames()
            list = e.GetFacenames()

            list.sort()

            self.lb1 = wx.ListBox(self, -1, wx.DefaultPosition, (200, 250),
                                 list, wx.LB_SINGLE)

            self.Bind(wx.EVT_LISTBOX, self.OnSelect, id=self.lb1.GetId())

            self.txt = wx.StaticText(self, -1, "Sample text...", (285, 50))

            sizer = wx.BoxSizer(wx.VERTICAL)
            sizer.Add(self.txt, 0, wx.EXPAND)
            sizer.Add(self.lb1, 0, wx.EXPAND|wx.TOP, 20)

            self.SetSizer(sizer)
            self.Fit()
            self.Layout()
            log.debug(self.GetSize())

            self.lb1.SetSelection(0)
            self.OnSelect(None)
            wx.FutureCall(300, self.SetTextSize)

        def SetTextSize(self):
            self.txt.SetSize(self.txt.GetBestSize())

        def OnSelect(self, evt):
            face = self.lb1.GetStringSelection()
            font = wx.Font(28, wx.DEFAULT, wx.NORMAL, wx.NORMAL, False, face)
            self.txt.SetLabel(face)
            self.txt.SetFont(font)
            if wx.Platform == "__WXMAC__": self.Refresh()

    def ButtonCB(parent, task, **kwargs):
        button = GenToggleButton(parent, -1, "Whatevar!!!")

    def CalendarCB(parent, task, **kwargs):
        wx.calendar.CalendarCtrl(parent, -1, wx.DateTime_Now())

    class TestSTC(wx.stc.StyledTextCtrl):
        def __init__(self, *args, **kwargs):
            wx.stc.StyledTextCtrl.__init__(self, *args, **kwargs)
            self.Bind(wx.stc.EVT_STC_UPDATEUI, self.OnUpdateUI)

        def OnUpdateUI(self, evt):
            """Specific OnUpdateUI callback for those modes that use an actual
            STC for their edit window.
            
            Adds things like fold level and style display.
            """
            linenum = self.GetCurrentLine()
            pos = self.GetCurrentPos()
            col = self.GetColumn(pos)
            status = "Line: %d Column: %d Position: %d" % (linenum, col, pos)
            if col == 0:
                status = ""
            self.GetParent().status.show_status_text(status)
            evt.Skip()

    app = wx.PySimpleApp()
    frm = wx.Frame(None,-1,"Test",style=wx.TAB_TRAVERSAL|wx.DEFAULT_FRAME_STYLE,
                   size=(600,400))
    panel = wx.Panel(frm)
    sizer = wx.BoxSizer(wx.HORIZONTAL)

    # spring tabs for the left side
    tabs1 = SpringTabs(panel, None)
    tabs1.add_tab("Calendar", CalendarCB)
    tabs1.add_tab("Fonts", FontList)
    tabs1.add_tab("Three", ButtonCB)
    sizer.Add(tabs1, 0, wx.EXPAND)

    text = TestSTC(panel, -1)
    text.SetText("Just a placeholder here.\nThe real action is on the borders\nand in the popup status bar.\n\nWhen the text cursor is in any column\nother than zero, the popup status\nbar will show the location.")
    sizer.Add(text, 1, wx.EXPAND)

    panel.status = PopupStatusBar(text)

    # spring tabs for the rigth side
    tabs2 = SpringTabs(panel, None, popup_direction="left")
    tabs2.add_tab("Calendar", CalendarCB)
    tabs2.add_tab("Five", ButtonCB)
    tabs2.add_tab("Six", ButtonCB)
    sizer.Add(tabs2, 0, wx.EXPAND)

    def fixFocus(evt):
        evt.Skip()
    text.Bind(wx.EVT_SET_FOCUS, fixFocus)

    panel.SetAutoLayout(True)
    panel.SetSizer(sizer)
    #sizer.Fit(panel)
    #sizer.SetSizeHints(panel)
    panel.Layout()
    app.SetTopWindow(frm)
    frm.Show()
    app.MainLoop()



"""
kml_templates

various templates for writting GNOME KML files
"""


LeStyles = """
    <Style id="RedDotIcon">
      <IconStyle>
         <scale>0.2</scale>
         <color>ff0000ff</color>
         <Icon>
            <href>dot.png</href>
         </Icon>
          <hotSpot x="0.5"  y="0.5" xunits="fraction" yunits="fraction"/>  
      </IconStyle>
      <LabelStyle>
         <color>00000000</color>
      </LabelStyle>
    </Style>

    <Style id="BlackDotIcon">
      <IconStyle>
         <scale>0.2</scale>
         <Icon>
            <href>dot.png</href>
         </Icon>
         <color>ff000000</color>
         <hotSpot x="0.5"  y="0.5" xunits="fraction" yunits="fraction"/>  
      </IconStyle>
      <LabelStyle>
         <color>00000000</color>
      </LabelStyle>
    </Style>
    
    <Style id="YellowDotIcon">
      <IconStyle>
         <scale>0.2</scale>
         <Icon>
            <href>dot.png</href>
         </Icon>
         <color>ff00efef</color>
         <hotSpot x="0.5"  y="0.5" xunits="fraction" yunits="fraction"/>  
      </IconStyle>
      <LabelStyle>
         <color>00000000</color>
      </LabelStyle>
    </Style>
 
    <Style id="RedXIcon">
      <IconStyle>
         <scale>0.2</scale>
         <color>ff0000ff</color>
         <Icon>
            <href>x.png</href>
         </Icon>
          <hotSpot x="0.5"  y="0.5" xunits="fraction" yunits="fraction"/>  
      </IconStyle>
      <LabelStyle>
         <color>00000000</color>
      </LabelStyle>
    </Style>

    <Style id="BlackXIcon">
      <IconStyle>
         <scale>0.2</scale>
         <Icon>
            <href>x.png</href>
         </Icon>
         <color>ff000000</color>
         <hotSpot x="0.5"  y="0.5" xunits="fraction" yunits="fraction"/>  
      </IconStyle>
      <LabelStyle>
         <color>00000000</color>
      </LabelStyle>
    </Style>
    
    <Style id="YellowXIcon">
      <IconStyle>
         <scale>0.2</scale>
         <Icon>
            <href>x.png</href>
         </Icon>
         <color>ff00efef</color>
         <hotSpot x="0.5"  y="0.5" xunits="fraction" yunits="fraction"/>  
      </IconStyle>
      <LabelStyle>
         <color>00000000</color>
      </LabelStyle>
    </Style>

"""


ContourStyles = """
    
    <Style id="HeavyStyle">
        <LineStyle><color>ff9E0000</color><width>1</width></LineStyle>  
        <PolyStyle><fill>1</fill><color>e09E0000</color></PolyStyle>
    </Style>
    
    <Style id="MediumStyle">
        <LineStyle><color>ffff7326</color><width>1</width></LineStyle>  
        <PolyStyle><fill>1</fill><color>e0ff7326</color></PolyStyle>
    </Style>
    
    <Style id="LightStyle">
        <LineStyle><color>ffffff00</color><width>1</width></LineStyle>  
        <PolyStyle><fill>1</fill><color>e0ffff00</color></PolyStyle>
    </Style>
    
    <Style id="UncertaintyStyle">
        <LineStyle><color>e0dcdcdc</color><width>2</width></LineStyle>  
        <PolyStyle><fill>1</fill><color>e0dcdcdc</color></PolyStyle>
    </Style>
"""

SaveOrigUncertainty = """
    <Style id="UncertaintyStyle">
        <LineStyle><color>ffff00ff</color><width>2</width></LineStyle>  
        <PolyStyle><fill>0</fill><color>e0ff00ff</color></PolyStyle>
    </Style>

"""

LogoOverlay = """
    % if fromLogo != "" :
    <Folder><name>Screen Overlays</name>
    <ScreenOverlay> 
    <name>Logo</name> 
    <color>ffffffff</color> 
    <visibility>1</visibility> 
    <Icon> 
    <href>${fromLogo}</href> 
    </Icon> 
    <overlayXY x="1" y="1" xunits="fraction" yunits="fraction"/> 
    <screenXY x="1" y="1" xunits="fraction" yunits="fraction"/> 
    <size x="-1" y="-1" xunits="fraction" yunits="fraction"/> 
    </ScreenOverlay> 
    </Folder>
    % endif
    """
    
    
# These icons (these are base64 encoded 3-pixel sized dots in a 32x32 transparent PNG)    
DotData="iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAAJOgAACToB8GSSSgAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAAEASURBVFiF7ZY7DsIwEEQfET09Ej11lFtwK06Re3ANlCoFPQpnoGJoHClCXpOPg10wUhonnnlyvF5vJJFSRdL0P0AOANsZcwqgAkrg6MZuQANcgdckN0ljn52kWlInW537ZjfWd2z4SVIbCP5U6+ZEAThLek4I7/V0cxcBnGaGDyGCK/Htn09ZdkutAnsiBFBHCO9VWzkb+XtBAdyB/Ywy9ekBHPCUqHUQVRHDcV6V74UFUEYMD3paAEdjfIm8nsl7gQVwWyHL62kBNCsAeD2zLcMXcIkUjvPyt+nASZj8KE7ejLJox1lcSIZ7IvqVzCrDkKJeSucARFW2veAP8DO9AXV74Qmb/4vgAAAAAElFTkSuQmCC"
XData="iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAAN1wAADdcBQiibeAAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAAHKSURBVFiFrdXLq01hGMfx12HMzMCUU4zFQEYiROYkEkkpHbeTXI5LSDqHtomBEJGY+RMMGBlKKWVmaiDXzvExsN7admuv9azLU89k7ef5fb/ruhMSVuIy3uEOVhXH++w1mMEbnMFSpITl+Ob/+oOpHuFHiszh+oIVCbPGVx8Sh0vguaYT3lcIdJU4VAGHtwm3agTaShysgcMgYUNAoKnEgQAcVueFqR4l9mMhkHVJ8RbkPt6DxL4g/EreGQ3oIrE3CL86vFd2FidaSOzBfGDn+ihv3KU82UBidxB+o4xV9TBFJSKX/eY4Tt0TfSooUVWzVYzIO326A3yuLj/6YWkjcTuSHRVImG4AH0RzJ1K8PqSUFoKzn8KpQdNd+N3wFoT+OyLwnfjVEB6WqIPv6AAPSVTBt+NnR3itxDj4tiD8Hs52kSiDb8WPQOB9LCp2WkuMwrcE4Q8xMbJ7ro3EcMBmfA8EPCqBt5bIi5uC8McV8Nznm0gkLMPXwMKTADz3haDExoRjgcGnWByEN5EYJLyuGXrWAp57pib7Y8K1ioHnHeC5L1bkP0iYHPPjCyzpCK+SmMdkHliLl8XBVzjaIzz3Ov++H59xF+uR/gJmOo2+fdNArAAAAABJRU5ErkJggg=="

################################
# main Template
################################

LE_Template = """<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2">
  <Document>
    <name>${name}</name>
    <open>1</open>
      % if description :
      <description><![CDATA[${description}]]></description>
      % endif
      
""" + LeStyles + """
    
% for seriesType, series in [("Best Guess",BestGuess_Series),("Uncertainty",Uncertainty_Series)]:
% if series and len(series) > 0 :
 <Folder>
 <name>${seriesType}</name>
% for seriesItem in series:
% for LEs,LEsname,LEsstyle,altitude in seriesItem.setsOfObjects:
   % if LEs :
     <Placemark>
      <name>${LEsname}</name>
      <styleUrl>${LEsstyle}</styleUrl>
      %if seriesItem.beginTime or seriesItem.endTime:
      <TimeSpan id="ID">
          % if seriesItem.beginTime :
          <begin>${seriesItem.beginTime}</begin>     <!-- kml:dateTime -->
          % endif
          % if seriesItem.endTime :
          <end>${seriesItem.endTime}</end>         <!-- kml:dateTime -->
          % endif
      </TimeSpan>
      % endif
      <MultiGeometry>
        % for le in LEs:
             <Point>
                 % if altitude == 0 :
                     <coordinates>${le.longitudeStr},${le.latitudeStr}</coordinates>
                 % else :
                     <altitudeMode>relativeToGround</altitudeMode>
                     <coordinates>${le.longitudeStr},${le.latitudeStr},${altitude}</coordinates>
                 % endif
             </Point>
        % endfor
        
      </MultiGeometry>
     </Placemark>
    %endif
% endfor
% endfor
 </Folder>
% endif
% endfor

""" + (
   LogoOverlay
)  +  """  
    
  </Document>
</kml>
"""






from __future__ import absolute_import

from traits.testing.unittest_tools import unittest

from ..heading_text import HeadingText
from ..split_dialog import SplitDialog
from ..toolkit import toolkit_object

GuiTestAssistant = toolkit_object('util.gui_test_assistant:GuiTestAssistant')
no_gui_test_assistant = (GuiTestAssistant.__name__ == 'Unimplemented')


@unittest.skipIf(no_gui_test_assistant, 'No GuiTestAssistant')
class TestDialog(unittest.TestCase, GuiTestAssistant):

    def setUp(self):
        GuiTestAssistant.setUp(self)
        self.dialog = SplitDialog()

    def tearDown(self):
        if self.dialog.control is not None:
            with self.delete_widget(self.dialog.control):
                self.dialog.destroy()
        GuiTestAssistant.tearDown(self)

    def test_create(self):
        # test that creation and destruction works as expected
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()

    def test_destroy(self):
        # test that destroy works even when no control
        self.dialog.destroy()

    def test_horizontal(self):
        # test that horizontal split works
        self.dialog.direction = 'horizontal'
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()

    def test_ratio(self):
        # test that ratio works
        self.dialog.ratio = 0.25
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()

    def test_contents(self):
        # test that contents works
        self.dialog.lhs = HeadingText
        self.dialog.rhs = HeadingText
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()



# -*- coding: ISO-8859-15 -*-
# =============================================================================
# Copyright (c) 2004, 2006 Sean C. Gillies
# Copyright (c) 2009 STFC <http://www.stfc.ac.uk>
#
# Authors :
#          Dominic Lowe <dominic.lowe@stfc.ac.uk>
#
# Contact email: dominic.lowe@stfc.ac.uk
# =============================================================================

"""
Web Map Server (WFS) methods and metadata. Factory function.
"""

from __future__ import (absolute_import, division, print_function)

from owslib.map import wms111, wms130


def WebMapService(url, version='1.1.1', xml=None, parse_remote_metadata=False,
                  timeout=30):

    '''wms factory function, returns a version specific WebMapService object

    @type url: string
    @param url: url of WFS capabilities document
    @type xml: string
    @param xml: elementtree object
    @type parse_remote_metadata: boolean
    @param parse_remote_metadata: whether to fully process MetadataURL elements
    @param timeout: time (in seconds) after which requests should timeout
    @return: initialized WebFeatureService_2_0_0 object
    '''
    if version in ['1.1.1']:
        return wms111.WebMapService_1_1_1(url, version, xml, parse_remote_metadata,
                                          timeout=timeout)
    elif version in ['1.3.0']:
        return wms130.WebMapService_1_3_0(url, version, xml, parse_remote_metadata,
                                          timeout=timeout)
    raise NotImplementedError('The WMS version (%s) you requested is not implemented. Please use 1.1.1 or 1.3.0.' % version)



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from pyface.api import error
from traits.api import Dict, Instance
from traitsui.api import Group, Handler, Item, SetEditor, View
from traitsui.menu import Action, CancelButton

# Local imports.
from apptools.permissions.package_globals import get_permissions_manager
from i_policy_storage import PolicyStorageError
from policy_data import Assignment, Role


class _AssignmentView(View):
    """The view for handling role assignments."""

    #### 'View' interface #####################################################

    kind = 'modal'

    title = "Assign roles"

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, all_roles, **traits):
        """Initialise the object."""

        buttons = [Action(name="Search"), Action(name="Save"), CancelButton]

        roles_editor = SetEditor(values=all_roles.values(),
                left_column_title="Available Roles",
                right_column_title="Assigned Roles")

        roles_group = Group(Item(name='roles', editor=roles_editor),
                label='Roles', show_border=True, show_labels=False)

        super(_AssignmentView, self).__init__(Item(name='user_name'),
                Item(name='description', style='readonly'), roles_group,
                buttons=buttons, **traits)


class _AssignmentHandler(Handler):
    """The view handler for role assignments."""

    #### Private interface ####################################################

    all_roles = Dict

    ###########################################################################
    # Trait handlers.
    ###########################################################################

    def _search_clicked(self, info):
        """Invoked by the "Search" button."""

        pm = get_permissions_manager()
        assignment = self._assignment(info)

        user = pm.user_manager.matching_user(assignment.user_name)
        if user is None:
            return

        try:
            user_name, role_names = pm.policy_manager.policy_storage.get_assignment(user.name)
        except PolicyStorageError, e:
            self._ps_error(e)
            return

        # Update the viewed object.
        assignment.user_name = user.name
        assignment.description = user.description
        assignment.roles = self._roles_to_list(role_names)

    def _save_clicked(self, info):
        """Invoked by the "Save" button."""

        assignment = self._validate(info)
        if assignment is None:
            return

        # Update the data in the database.
        try:
            get_permissions_manager().policy_manager.policy_storage.set_assignment(assignment.user_name, [r.name for r in assignment.roles])

            info.ui.dispose()
        except PolicyStorageError, e:
            self._ps_error(e)

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _validate(self, info):
        """Validate the assignment and return it if there were no problems."""

        assignment = self._assignment(info)
        assignment.user_name = assignment.user_name.strip()

        if not assignment.user_name:
            self._error("A user name must be given.")
            return None

        return assignment

    def _roles_to_list(self, role_names):
        """Return a list of Role instances created from the given list of role
        names."""

        rl = []

        for name in role_names:
            try:
                r = self.all_roles[name]
            except KeyError:
                # This shouldn't happen if the policy database is maintaining
                # referential integrity.
                continue

            rl.append(r)

        return rl

    @staticmethod
    def _assignment(info):
        """Return the assignment instance being handled."""

        return info.ui.context['object']

    @staticmethod
    def _error(msg):
        """Display an error message to the user."""

        error(None, msg)

    @staticmethod
    def _ps_error(e):
        """Display a message to the user after a PolicyStorageError exception
        has been raised."""

        error(None, str(e))


def role_assignment():
    """Implement the role assignment for the current policy manager."""

    # Create a dictionary of roles keyed by the role name.
    all_roles = {}

    try:
        roles = get_permissions_manager().policy_manager.policy_storage.all_roles()
    except PolicyStorageError, e:
        error(None, str(e))
        return

    for name, description in roles:
        all_roles[name] = Role(name=name, description=description)

    assignment = Assignment()
    view = _AssignmentView(all_roles)
    handler = _AssignmentHandler(all_roles=all_roles)

    assignment.edit_traits(view=view, handler=handler)



#------------------------------------------------------------------------------
# Copyright (c) 2008, Enthought, Inc.
# All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#
#------------------------------------------------------------------------------
""" Defines a drop editor factory for all traits toolkit backends.
    A drop target editor handles drag and drop operations as a drop
    target.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Any, Bool

from .text_editor import ToolkitEditorFactory as EditorFactory

#-------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------


class ToolkitEditorFactory(EditorFactory):
    """ Editor factory for drop editors.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Allowable drop objects must be of this class (optional)
    klass = Any

    # Must allowable drop objects be bindings?
    binding = Bool(False)

    # Can the user type into the editor, or is it read only?
    readonly = Bool(True)

# Define the DropEditor class.
DropEditor = ToolkitEditorFactory

### EOF #######################################################################



"""

  fs.tests.test_path:  testcases for the fs path functions

"""


import unittest
import fs.tests

from fs.path import *

class TestPathFunctions(unittest.TestCase):
    """Testcases for FS path functions."""

    def test_normpath(self):
        tests = [   ("\\a\\b\\c", "\\a\\b\\c"),
                    (".", ""),
                    ("./", ""),
                    ("", ""),
                    ("/.", "/"),
                    ("/a/b/c", "/a/b/c"),
                    ("a/b/c", "a/b/c"),
                    ("a/b/../c/", "a/c"),
                    ("/","/"),
                    (u"a/\N{GREEK SMALL LETTER BETA}/c",u"a/\N{GREEK SMALL LETTER BETA}/c"),
                    ]
        for path, result in tests:
            self.assertEqual(normpath(path), result)

    def test_pathjoin(self):
        tests = [   ("", "a", "a"),
                    ("a", "a", "a/a"),
                    ("a/b", "../c", "a/c"),
                    ("a/b/../c", "d", "a/c/d"),
                    ("/a/b/c", "d", "/a/b/c/d"),
                    ("/a/b/c", "../../../d", "/d"),
                    ("a", "b", "c", "a/b/c"),
                    ("a/b/c", "../d", "c", "a/b/d/c"),
                    ("a/b/c", "../d", "/a", "/a"),
                    ("aaa", "bbb/ccc", "aaa/bbb/ccc"),
                    ("aaa", "bbb\\ccc", "aaa/bbb\\ccc"),
                    ("aaa", "bbb", "ccc", "/aaa", "eee", "/aaa/eee"),
                    ("a/b", "./d", "e", "a/b/d/e"),
                    ("/", "/", "/"),
                    ("/", "", "/"),
                    (u"a/\N{GREEK SMALL LETTER BETA}","c",u"a/\N{GREEK SMALL LETTER BETA}/c"),
        ]
        for testpaths in tests:
            paths = testpaths[:-1]
            result = testpaths[-1]
            self.assertEqual(pathjoin(*paths), result)

        self.assertRaises(ValueError, pathjoin, "..")
        self.assertRaises(ValueError, pathjoin, "../")
        self.assertRaises(ValueError, pathjoin, "/..")
        self.assertRaises(ValueError, pathjoin, "./../")
        self.assertRaises(ValueError, pathjoin, "a/b", "../../..")
        self.assertRaises(ValueError, pathjoin, "a/b/../../../d")

    def test_relpath(self):
        tests = [   ("/a/b", "a/b"),
                    ("a/b", "a/b"),
                    ("/", "") ]

        for path, result in tests:
            self.assertEqual(relpath(path), result)

    def test_abspath(self):
        tests = [   ("/a/b", "/a/b"),
                    ("a/b", "/a/b"),
                    ("/", "/") ]

        for path, result in tests:
            self.assertEqual(abspath(path), result)

    def test_iteratepath(self):
        tests = [   ("a/b", ["a", "b"]),
                    ("", [] ),
                    ("aaa/bbb/ccc", ["aaa", "bbb", "ccc"]),
                    ("a/b/c/../d", ["a", "b", "d"]) ]

        for path, results in tests:
            for path_component, expected in zip(iteratepath(path), results):
                self.assertEqual(path_component, expected)

        self.assertEqual(list(iteratepath("a/b/c/d", 1)), ["a", "b/c/d"])
        self.assertEqual(list(iteratepath("a/b/c/d", 2)), ["a", "b", "c/d"])

    def test_pathsplit(self):
        tests = [   ("a/b", ("a", "b")),
                    ("a/b/c", ("a/b", "c")),
                    ("a", ("", "a")),
                    ("", ("", "")),
                    ("/", ("/", "")),
                    ("/foo", ("/", "foo")),
                    ("foo/bar", ("foo", "bar")),
                    ("foo/bar/baz", ("foo/bar", "baz")),
                ]
        for path, result in tests:
            self.assertEqual(pathsplit(path), result)

    def test_recursepath(self):
        self.assertEquals(recursepath("/"),["/"])
        self.assertEquals(recursepath("hello"),["/","/hello"])
        self.assertEquals(recursepath("/hello/world/"),["/","/hello","/hello/world"])
        self.assertEquals(recursepath("/hello/world/",reverse=True),["/hello/world","/hello","/"])
        self.assertEquals(recursepath("hello",reverse=True),["/hello","/"])
        self.assertEquals(recursepath("",reverse=True),["/"])

    def test_isdotfile(self):
        for path in ['.foo',
                     '.svn',
                     'foo/.svn',
                     'foo/bar/.svn',
                     '/foo/.bar']:
            self.assert_(isdotfile(path))

        for path in ['asfoo',
                     'df.svn',
                     'foo/er.svn',
                     'foo/bar/test.txt',
                     '/foo/bar']:
            self.assertFalse(isdotfile(path))

    def test_dirname(self):
        tests = [('foo', ''),
                 ('foo/bar', 'foo'),
                 ('foo/bar/baz', 'foo/bar'),
                 ('/foo/bar', '/foo'),
                 ('/foo', '/'),
                 ('/', '/')]
        for path, test_dirname in tests:
            self.assertEqual(dirname(path), test_dirname)

    def test_basename(self):
        tests = [('foo', 'foo'),
                 ('foo/bar', 'bar'),
                 ('foo/bar/baz', 'baz'),
                 ('/', '')]
        for path, test_basename in tests:
            self.assertEqual(basename(path), test_basename)

    def test_iswildcard(self):
        self.assert_(iswildcard('*'))
        self.assert_(iswildcard('*.jpg'))
        self.assert_(iswildcard('foo/*'))
        self.assert_(iswildcard('foo/{}'))
        self.assertFalse(iswildcard('foo'))
        self.assertFalse(iswildcard('img.jpg'))
        self.assertFalse(iswildcard('foo/bar'))

    def test_realtivefrom(self):
        tests = [('/', '/foo.html', 'foo.html'),
                 ('/foo', '/foo/bar.html', 'bar.html'),
                 ('/foo/bar/', '/egg.html', '../../egg.html'),
                 ('/a/b/c/d', 'e', '../../../../e'),
                 ('/a/b/c/d', 'a/d', '../../../d'),
                 ('/docs/', 'tags/index.html', '../tags/index.html'),
                 ('foo/bar', 'baz/index.html', '../../baz/index.html'),
                 ('', 'a', 'a'),
                 ('a', 'b/c', '../b/c')
                 ]

        for base, path, result in tests:
            self.assertEqual(relativefrom(base, path), result)


class Test_PathMap(unittest.TestCase):

    def test_basics(self):
        map = PathMap()
        map["hello"] = "world"
        self.assertEquals(map["/hello"],"world")
        self.assertEquals(map["/hello/"],"world")
        self.assertEquals(map.get("hello"),"world")

    def test_iteration(self):
        map = PathMap()
        map["hello/world"] = 1
        map["hello/world/howareya"] = 2
        map["hello/world/iamfine"] = 3
        map["hello/kitty"] = 4
        map["hello/kitty/islame"] = 5
        map["batman/isawesome"] = 6
        self.assertEquals(set(map.iterkeys()),set(("/hello/world","/hello/world/howareya","/hello/world/iamfine","/hello/kitty","/hello/kitty/islame","/batman/isawesome")))
        self.assertEquals(sorted(map.values()),range(1,7))
        self.assertEquals(sorted(map.items("/hello/world/")),[("/hello/world",1),("/hello/world/howareya",2),("/hello/world/iamfine",3)])
        self.assertEquals(zip(map.keys(),map.values()),map.items())
        self.assertEquals(zip(map.keys("batman"),map.values("batman")),map.items("batman"))
        self.assertEquals(set(map.iternames("hello")),set(("world","kitty")))
        self.assertEquals(set(map.iternames("/hello/kitty")),set(("islame",)))

        del map["hello/kitty/islame"]
        self.assertEquals(set(map.iternames("/hello/kitty")),set())
        self.assertEquals(set(map.iterkeys()),set(("/hello/world","/hello/world/howareya","/hello/world/iamfine","/hello/kitty","/batman/isawesome")))
        self.assertEquals(set(map.values()),set(range(1,7)) - set((5,)))





#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms described in the PyQt GPL exception also apply

#
# Author: Riverbank Computing Limited
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------


# Standard library imports.
import sys
from os.path import basename

# Major package imports.
import wx
import wx.stc

# Enthought library imports.
from traits.api import Bool, Event, Instance, File, Unicode, Property, provides
from pyface.wx.python_stc import PythonSTC, faces

# Local imports.
from omnivore.framework.editor import FrameworkEditor
from i_styled_text_editor import IStyledTextEditor
from pyface.key_pressed_event import KeyPressedEvent


@provides(IStyledTextEditor)
class StyledTextEditor(FrameworkEditor):
    """ The toolkit specific implementation of a StyledTextEditor.  See the
    IStyledTextEditor interface for the API documentation.
    """

    #### 'IPythonEditor' interface ############################################

    obj = Instance(File)

    show_line_numbers = Bool(True)

    #### Events ####

    changed = Event

    key_pressed = Event(KeyPressedEvent)

    ###########################################################################
    # 'PythonEditor' interface.
    ###########################################################################

    def create(self, parent):
        self.control = self._create_control(parent)

    def rebuild_document_properties(self):
        text = self.document.to_bytes()
        self.control.SetTextUTF8(text)
        self.control.EmptyUndoBuffer()

    def undo(self):
        self.control.Undo()

    def redo(self):
        self.control.Redo()

    def select_line(self, lineno):
        """ Selects the specified line.
        """
        start = self.control.PositionFromLine(lineno)
        end   = self.control.GetLineEndPosition(lineno)

        self.control.SetSelection(start, end)

    def set_style(self, n, fore, back):
        self.control.StyleSetForeground(n, fore)
        #self.StyleSetBackground(n, '#c0c0c0')
        #self.StyleSetBackground(n, '#ffffff')
        self.control.StyleSetBackground(n, back)
        self.control.StyleSetFaceName(n, "courier new")
        self.control.StyleSetSize(n, faces['size'])

    ###########################################################################
    # Trait handlers.
    ###########################################################################

    def _show_line_numbers_changed(self):
        if self.control is not None:
            c = self.control
            if self.show_line_numbers:
                c.SetMarginType(1, wx.stc.STC_MARGIN_NUMBER)
                c.SetMarginWidth(1, 45)
            else:
                c.SetMarginWidth(1, 4)

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_control(self, parent):
        """ Creates the toolkit-specific control for the widget.
        """
        from pyface.ui.qt4.code_editor.code_widget import AdvancedCodeWidget
        self.control = control = AdvancedCodeWidget(parent)
        self._show_line_numbers_changed()

    def _create_control(self, parent):
        """ Creates the toolkit-specific control for the widget. """

        # Base-class constructor.
        self.control = stc = PythonSTC(parent, -1)

        # No folding!
        stc.SetProperty("fold", "0")

        # Mark the maximum line size.
        stc.SetEdgeMode(wx.stc.STC_EDGE_LINE)
        stc.SetEdgeColumn(79)

        stc.SetMarginType(1, wx.stc.STC_MARGIN_NUMBER)
        stc.SetMarginMask(1, ~wx.stc.STC_MASK_FOLDERS)
        # Display line numbers in the margin.
        if self.show_line_numbers:
            stc.SetMarginWidth(1, 45)
            self.set_style(wx.stc.STC_STYLE_LINENUMBER, "#000000", "#c0c0c0")
        else:
            stc.SetMarginWidth(1, 4)
            self.set_style(wx.stc.STC_STYLE_LINENUMBER, "#ffffff", "#ffffff")

        # Create 'tabs' out of spaces!
        stc.SetUseTabs(False)

        # One 'tab' is 4 spaces.
        stc.SetIndent(4)

        # Line ending mode.
        stc.SetEOLMode(wx.stc.STC_EOL_LF) # Unix
        #self.SetEOLMode(wx.stc.STC_EOL_CR) # Apple Mac
        #self.SetEOLMode(wx.stc.STC_EOL_CRLF) # Windows

        ##########################################
        # Global styles for all languages.
        ##########################################

        self.set_style(wx.stc.STC_STYLE_DEFAULT, "#000000", "#ffffff")
        self.set_style(wx.stc.STC_STYLE_CONTROLCHAR, "#000000", "#ffffff")
        self.set_style(wx.stc.STC_STYLE_BRACELIGHT, "#000000", "#ffffff")
        self.set_style(wx.stc.STC_STYLE_BRACEBAD, "#000000", "#ffffff")

        ##########################################
        # Python styles.
        ##########################################

        # White space
        self.set_style(wx.stc.STC_P_DEFAULT, "#000000", "#ffffff")

        # Comment
        self.set_style(wx.stc.STC_P_COMMENTLINE, "#007f00", "#ffffff")

        # Number
        self.set_style(wx.stc.STC_P_NUMBER, "#007f7f", "#ffffff")

        # String
        self.set_style(wx.stc.STC_P_STRING, "#7f007f", "#ffffff")

        # Single quoted string
        self.set_style(wx.stc.STC_P_CHARACTER, "#7f007f", "#ffffff")

        # Keyword
        self.set_style(wx.stc.STC_P_WORD, "#00007f", "#ffffff")

        # Triple quotes
        self.set_style(wx.stc.STC_P_TRIPLE, "#7f0000", "#ffffff")

        # Triple double quotes
        self.set_style(wx.stc.STC_P_TRIPLEDOUBLE, "#ff0000", "#ffffff")

        # Class name definition
        self.set_style(wx.stc.STC_P_CLASSNAME, "#0000ff", "#ffffff")

        # Function or method name definition
        self.set_style(wx.stc.STC_P_DEFNAME, "#007f7f", "#ffffff")

        # Operators
        self.set_style(wx.stc.STC_P_OPERATOR, "#000000", "#ffffff")

        # Identifiers
        self.set_style(wx.stc.STC_P_IDENTIFIER, "#000000", "#ffffff")

        # Comment-blocks
        self.set_style(wx.stc.STC_P_COMMENTBLOCK, "#007f00", "#ffffff")

        # End of line where string is not closed
        self.set_style(wx.stc.STC_P_STRINGEOL, "#000000", "#ffffff")

        ##########################################
        # Events.
        ##########################################

        # By default, the will fire EVT_STC_CHANGE evented for all mask values
        # (STC_MODEVENTMASKALL). This generates too many events.
        stc.SetModEventMask(wx.stc.STC_MOD_INSERTTEXT |
                            wx.stc.STC_MOD_DELETETEXT |
                            wx.stc.STC_PERFORMED_UNDO |
                            wx.stc.STC_PERFORMED_REDO)

        # Listen for changes to the file.
        wx.stc.EVT_STC_CHANGE(stc, stc.GetId(), self._on_stc_changed)

        # Listen for key press events.
        wx.EVT_CHAR(stc, self._on_char)

        # Load the editor's contents.
        self.load()

        return stc

    #### wx event handlers ####################################################

    def _on_stc_changed(self, event):
        """ Called whenever a change is made to the text of the document. """

        self.dirty = self.control.CanUndo()
        self.can_undo = self.control.CanUndo()
        self.can_redo = self.control.CanRedo()
        self.changed = True

        # Give other event handlers a chance.
        event.Skip()

        return

    def _on_char(self, event):
        """ Called whenever a change is made to the text of the document. """

        self.key_pressed = KeyPressedEvent(
            alt_down     = event.m_altDown == 1,
            control_down = event.m_controlDown == 1,
            shift_down   = event.m_shiftDown == 1,
            key_code     = event.m_keyCode,
            event        = event
        )

        # Give other event handlers a chance.
        event.Skip()

        return



from __future__ import absolute_import

import os

from traits.etsconfig.api import ETSConfig
from traits.testing.unittest_tools import unittest

from ..message_dialog import MessageDialog, information, warning, error
from ..constant import OK
from ..gui import GUI
from ..toolkit import toolkit_object
from ..window import Window


GuiTestAssistant = toolkit_object('util.gui_test_assistant:GuiTestAssistant')
no_gui_test_assistant = (GuiTestAssistant.__name__ == 'Unimplemented')

ModalDialogTester = toolkit_object('util.modal_dialog_tester:ModalDialogTester')
no_modal_dialog_tester = (ModalDialogTester.__name__ == 'Unimplemented')

USING_QT = ETSConfig.toolkit not in ['', 'null', 'wx']
is_pyqt5 = (ETSConfig.toolkit == 'qt4' and os.environ.get('QT_API') == 'pyqt5')


@unittest.skipIf(no_gui_test_assistant, 'No GuiTestAssistant')
class TestMessageDialog(unittest.TestCase, GuiTestAssistant):

    def setUp(self):
        GuiTestAssistant.setUp(self)
        self.dialog = MessageDialog()

    def tearDown(self):
        if self.dialog.control is not None:
            with self.delete_widget(self.dialog.control):
                self.dialog.destroy()
        GuiTestAssistant.tearDown(self)

    def test_create(self):
        # test that creation and destruction works as expected
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()

    def test_destroy(self):
        # test that destroy works even when no control
        self.dialog.destroy()

    def test_size(self):
        # test that size works as expected
        self.dialog.size = (100, 100)
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()

    def test_position(self):
        # test that position works as expected
        self.dialog.position = (100, 100)
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()

    def test_create_parent(self):
        # test that creation and destruction works as expected with a parent
        parent = Window()
        self.dialog.parent = parent.control
        parent._create()
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()
        parent.destroy()

    def test_create_ok_renamed(self):
        # test that creation and destruction works as expected with ok_label
        self.dialog.ok_label = u"Sure"
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()

    def test_message(self):
        # test that creation and destruction works as expected with message
        self.dialog.message = u"This is the message"
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()

    def test_informative(self):
        # test that creation and destruction works with informative
        self.dialog.message = u"This is the message"
        self.dialog.informative = u"This is the additional message"
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()

    def test_detail(self):
        # test that creation and destruction works with detail
        self.dialog.message = u"This is the message"
        self.dialog.informative = u"This is the additional message"
        self.dialog.detail = u"This is the detail"
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()

    def test_warning(self):
        # test that creation and destruction works with warning message
        self.dialog.severity = "warning"
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()

    def test_error(self):
        # test that creation and destruction works with error message
        self.dialog.severity = "error"
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_accept(self):
        # test that accept works as expected
        # XXX duplicate of Dialog test, not needed?
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_run(when_opened=lambda x: x.close(accept=True))
        self.assertEqual(tester.result, OK)
        self.assertEqual(self.dialog.return_code, OK)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_close(self):
        # test that closing works as expected
        # XXX duplicate of Dialog test, not needed?
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_run(when_opened=lambda x: self.dialog.close())
        self.assertEqual(tester.result, OK)
        self.assertEqual(self.dialog.return_code, OK)

    @unittest.skipIf(is_pyqt5, "Message dialog click tests don't work on pyqt5.")  # noqa
    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_ok(self):
        # test that OK works as expected
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_wait(when_opened=lambda x: x.click_button(OK))
        self.assertEqual(tester.result, OK)
        self.assertEqual(self.dialog.return_code, OK)

    @unittest.skipIf(USING_QT, "Can't change OK label in Qt")
    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_renamed_ok(self):
        self.dialog.ok_label = u"Sure"
        # test that OK works as expected if renamed
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_wait(when_opened=lambda x: x.click_widget(u"Sure"))
        self.assertEqual(tester.result, OK)
        self.assertEqual(self.dialog.return_code, OK)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_parent(self):
        # test that lifecycle works with a parent
        parent = Window()
        self.dialog.parent = parent.control
        parent.open()
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_run(when_opened=lambda x: x.close(accept=True))
        parent.close()
        self.assertEqual(tester.result, OK)
        self.assertEqual(self.dialog.return_code, OK)


@unittest.skipIf(no_gui_test_assistant, 'No GuiTestAssistant')
@unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
class TestMessageDialogHelpers(unittest.TestCase, GuiTestAssistant):

    def test_information(self):
        self._check_dialog(information)

    def test_warning(self):
        self._check_dialog(warning)

    def test_error(self):
        self._check_dialog(error)

    def _check_dialog(self, helper):
        message = 'message'
        kwargs = {
            'title': 'Title',
            'detail': 'Detail',
            'informative': 'Informative'
        }

        # smoke test, since dialog helper is opaque
        result = self._open_and_close(helper, message, **kwargs)

        self.assertIsNone(result)

    def _open_and_close(self, helper, message, **kwargs):
        parent = Window()
        parent.open()

        when_opened = lambda x: x.close(accept=True)

        tester = ModalDialogTester(helper)
        tester.open_and_wait(when_opened, parent.control, message, **kwargs)

        parent.close()
        return tester.result



"""
Tests for the utilities module.

"""

from __future__ import (absolute_import, division, print_function)

import numpy as np
from pyugrid import util


class DummyArrayLike(object):
    """
    Class that will look like an array to this function, even
    though it won't work!

    Just for tests. All it does is add a few expected attributes

    This will need to be updated when the function is changed.

    """
    must_have = ['dtype', 'shape', 'ndim', '__len__', '__getitem__', '__getattribute__']

    # pretty kludgy way to do this..
    def __new__(cls):
        print ("in new"), cls
        obj = object.__new__(cls)
        for attr in cls.must_have:
            setattr(obj, attr, None)
        return obj


def test_dummy_array_like():
    dum = DummyArrayLike()
    print(dum)
    print(dum.dtype)
    for attr in DummyArrayLike.must_have:
        assert hasattr(dum, attr)


def test_asarraylike_list():
    """
    Passing in a list should return a np.ndarray.

    """
    lst = [1, 2, 3, 4]
    result = util.asarraylike(lst)
    assert isinstance(result, np.ndarray)
    assert np.array_equal(result, lst)


def test_asarraylike_array():
    """
    Passing in a list should return a np.ndarray.

    """
    arr = np.array([1, 2, 3, 4])
    result = util.asarraylike(arr)

    assert result is arr


def test_as_test_asarraylike_dummy():
    dum = DummyArrayLike()
    result = util.asarraylike(dum)
    assert result is dum



import os
import sys

from omnivore.utils.textutil import slugify

import logging
log = logging.getLogger(__name__)

class OmnivoreDocumentationError(RuntimeError):
    pass

class AlreadySeenError(OmnivoreDocumentationError):
    pass

class SkipDocumentationError(OmnivoreDocumentationError):
    pass

# The hierarchy coming from task.get_menu_action_hierarchy() is a list of
# tuples that look like this:
#
# ("Menubar -> ", None)
# ("Menubar -> File -> ", None)
# ("Menubar -> File -> New -> ", None)
# ("Menubar -> File -> New -> Bitmap File", <omnivore.framework.actions.NewFileAction object at 0x7f28ceeaad10>)
# ("Menubar -> File -> New -> Bitmap Image", <omnivore.framework.actions.NewFileAction object at 0x7f28cee8bfb0>)
# ("Menubar -> File -> New -> Text File", <omnivore.framework.actions.NewFileAction object at 0x7f28cee8be30>)
# ("Menubar -> File -> Open...", <omnivore.framework.actions.OpenAction object at 0x7f28cf75b590>)
# ("Menubar -> File -> Open Recent -> ", None)
# ("Menubar -> File -> Open Recent -> about://omnivore", <omnivore.plugins.open_recent.OpenRecentAction object at 0x7f28ceec4830>)
# ("Menubar -> File -> Insert File...", <omnivore8bit.hex_edit.actions.InsertFileAction object at 0x7f28cf75b830>)
# ("Menubar -> File -> Save", <omnivore.framework.actions.SaveAction object at 0x7f28cf75bd70>)
# ("Menubar -> File -> Save As...", <omnivore.framework.actions.SaveAsAction object at 0x7f28cf75bdd0>)
# ("Menubar -> File -> Save Segment As -> ", None)
# ("Menubar -> File -> Save As Image...", <omnivore.framework.actions.SaveAsImageAction object at 0x7f28cf781470>)
# ("Menubar -> File -> Revert", <omnivore.framework.actions.RevertAction object at 0x7f28cf781c50>)
# ("Menubar -> File -> Page Setup...", <omnivore.framework.actions.PageSetupAction object at 0x7f28cf71c1d0>)
# ("Menubar -> File -> Print Preview", <omnivore.framework.actions.PrintPreviewAction object at 0x7f28cf71c230>)
# ("Menubar -> File -> Print...", <omnivore.framework.actions.PrintAction object at 0x7f28cf71c290>)
# ("Menubar -> File -> Export as XEX...", <omnivore8bit.hex_edit.actions.SaveAsXEXAction object at 0x7f28cf71ccb0>)
# ("Menubar -> File -> Export as Boot Disk...", <omnivore8bit.hex_edit.actions.SaveAsXEXBootAction object at 0x7f28cf71cd10>)
# ("Menubar -> File -> Quit", <omnivore.framework.actions.ExitAction object at 0x7f28cf737650>)
# ("Menubar -> Edit -> ", None)
# ("Menubar -> Edit -> Undo", <omnivore.framework.actions.UndoAction object at 0x7f28cf737950>)
# ("Menubar -> Edit -> Redo", <omnivore.framework.actions.RedoAction object at 0x7f28cf7379b0>)
# ("Menubar -> Edit -> Revert to Baseline Data", <omnivore8bit.hex_edit.actions.RevertToBaselineAction object at 0x7f28cf737cb0>)

def split_path(path):
    parts = path.split(" -> ")
    title = parts[-1].rstrip("...")
    if title:
        menu = parts[0:-1]
        is_action = True
    else:  # skip the last one, it's blank
        title = parts[-2].rstrip("...")
        menu = parts[0:-1]
        is_action = False
    # print "SPLIT", parts
    return menu, title, len(menu), is_action

def trim(docstring):
    # docstring formatter from PEP-257
    if not docstring:
        return ''
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxint
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxint:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return '\n'.join(trimmed)

def get_best_doc(action):
    if action.__doc__:
        return trim(action.__doc__)
    else:
        return action.description or action.tooltip

def skip_action(action):
    name = action.__class__.__name__
    return name == "DockPaneToggleAction" or name == "TaskToggleAction"


rst_toc_entry_template = "   {0}"

rst_toc_of_subdir_template = "   {0}/index"

rst_section_chars = {
    2: "=",  # Menu items, first submenu
    3: "-",  # Submenu items, second submenu
    4: "~",
    5: "^"
}

def get_doc_hint(item):
    text = getattr(item, "doc_hint", "")
    hints = [t.strip() for t in text.split(",")]
    return hints

def get_rst_section_title(level, title, page=False, extra_docs=""):
    divider = rst_section_chars[level] * len(title)
    return "\n\n%s\n%s\n%s\n\n%s\n\n" % (divider if page else "", title, divider, extra_docs)

def get_rst_action_description(level, title, action_info, in_submenu=True):
    text = action_info[0]
    doc_hint = action_info[1]
    parent_doc = action_info[2]
    submenu_of = action_info[3]
    extra_indent = action_info[4]
    lines = []
    indent = ""

    level += extra_indent
    if submenu_of:
        # The submenu label is on the children because the group of submenu
        # items is dynamically generated or something.
        lines.append(get_rst_section_title(level, submenu_of))

    if parent_doc:
        lines.append(parent_doc)
        lines.append("\n\n")
    if "summary" in doc_hint:
        # just use text as is because the menu title will have already been
        # printed
        level = -1
    if "parent" in doc_hint:
        if "list" in doc_hint:
            if "compact" in doc_hint:
                # compact list puts the item description on the same line as
                # the title
                lines.append("%s:" % title)
                indent = "    "
            elif "level 3" in doc_hint:
                # the submenu items in this list should be shown as RST
                # sections at level 3, below the level 2 of the parent menu.
                # This is used when the parent menu is dynamically generated
                # and there's no good place to put a docstring on the parent
                lines.append(get_rst_section_title(3, title))
            else:
                # normal list is for a menu that can't have a docstring on
                # their own because they aren't an action. The documentation is
                # on the submenu items and the docstring text of those items
                # refers to the parent menu. So the description needs to come
                # before the first item becasue it refers to the menu title,
                # then the title of the submenu item.
                text += "\n\n* %s" % title
    elif in_submenu:
        if level < 0:
            # do nothing, format text as is
            pass
        elif level == 2 or level == 3:  # Actions in the main pulldown are subsections
            lines.append(get_rst_section_title(level, title))
        elif level == 4:  # Actions in the first submenu level
            lines.append("%s:" % title)
            indent = "    "
        else:  # Actions in deeper submenus
            lines.append("* %s:" % title)
            text = ""  # force no description
    else:
        if level < 0:
            # do nothing, format text as is
            pass
        elif level == 2 or level == 3:  # Actions in the main pulldown are subsections
            lines.append(get_rst_section_title(level + 1, title))
        # elif level == 4:  # Actions in the first submenu level
        #     lines.append("%s:" % title)
        #     indent = "    "
        else:  # Actions in deeper submenus
            lines.append("* %s:" % title)
            text = ""  # force no description

    lines.extend([indent + t for t in text.splitlines()])
    lines.append("")
    return lines


class RSTSeparateMenuDocs(object):
    default_templates = {
        "manual_index": """
.. _{slug}

{title}

Introduction
============

{intro}

Editors
=======

.. toctree::
   :maxdepth: 2

{toc}
""",

        "task_index":"""
.. _{slug}

{title}

Overview
========

{overview}

Menus
=====

.. toctree::
   :maxdepth: 2

{toc}
""",

        "task_index_one_page":"""
.. _{slug}

{title}

.. toctree::
   :maxdepth: 2

{toc}

Overview
========

{overview}

----

{pages}
""",

        "page": """
.. _{slug}

{title}

Overview
********

Menu Items
**********
""",

        "page_one_page": """
.. _{slug}

{title}

{description}


""",
    }


    def __init__(self, title, directory):
        self.title = title
        self.title_slug = slugify(title)
        self.directory = directory
        self.template_dir = directory
        self.template_suffix = ".rst.in"
        self.sections = []

    def get_subdir(self, name):
        subdir = os.path.join(self.directory, name)
        try:
            os.mkdir(subdir)
        except OSError, e:
            # directory exists!
            pass
        return subdir

    def get_template(self, kind):
        filename = os.path.join(self.template_dir, kind + self.template_suffix)
        try:
            with open(filename, "r") as fh:
                template = fh.read()
        except IOError, e:
            template = self.default_templates[kind]
        return template

    def get_action_info(self, action, menu, summaries_seen):
        doc_hint = get_doc_hint(action)
        summary_id = "/".join(menu) + "/" + action.__class__.__name__
        text = None
        parent_doc = ""
        submenu_of = ""
        extra_indent = 0
        if "summary" in doc_hint:
            if summary_id in summaries_seen:
                raise AlreadySeenError
            summaries_seen.add(summary_id)
        if "ignore" in doc_hint:
            raise AlreadySeenError
        elif "parent" in doc_hint:
            # 'parent' keyword appears on actions where its parent menu is not
            # an action. This occurs on dynamically generated menus and almost
            # everything that is in a submenu. The first time, the description
            # of this will show up under its parent title, but siblings should
            # not duplicate this.
            if summary_id in summaries_seen:
                text = ""
            summaries_seen.add(summary_id)
        parent_doc = trim(getattr(action, "parent_doc", "")).format(name=action.name, tooltip=action.tooltip)
        if parent_doc:
            # Only on the first time the *parent* of this item is seen should
            # the parent_doc be displayed.
            summary_id = "/".join(menu)
            if summary_id in summaries_seen:
                parent_doc = ""
            summaries_seen.add(summary_id)
        submenu_of = trim(getattr(action, "submenu_of", "")).format(name=action.name, tooltip=action.tooltip)
        if submenu_of:
            # Only on the first time the *parent* of this item is seen should
            # the submenu label be displayed.
            summary_id = "/".join(menu) + "/" + submenu_of
            if summary_id in summaries_seen:
                submenu_of = ""
            summaries_seen.add(summary_id)

            # But, all the entries should be indented
            extra_indent = 1
        if "skip" in doc_hint:
            raise SkipDocumentationError
        if text is None:
            text = get_best_doc(action)
        text = text.format(name=action.name, tooltip=action.tooltip)
        return (text, doc_hint, parent_doc, submenu_of, extra_indent)

    def create_task_sections(self, directory, hierarchy, base_slug):
        toc_entries = []
        pages = []
        current_page = []
        summaries_seen = set()
        template = self.get_template("page")
        for path, action in hierarchy:
            if skip_action(action):
                continue
            menu, title, level, is_action = split_path(path)
            if level > 1:
                if not is_action:  # explicit menu
                    if level == 2:  # toplevel menu item
                        slug = "%s.%s" % (base_slug, slugify(title))
                        toc_entries.append((slug, title))
                        subs = {
                            "slug": slug,
                            "title": get_rst_section_title(level, title, True),
                        }
                        current_page = [template.format(**subs)]
                        print "New page for %s: %s" % (title, slug)
                        log.debug("New page for %s: %s" % (title, slug))
                        pages.append((slug, title, current_page))
                    else:
                        log.debug("Submenu %s")
                        current_page.append(get_rst_section_title(level - 1, title))

                else:  # menu item could be in a submenu or up a level
                    try:
                        action_info = self.get_action_info(action, menu, summaries_seen)
                    except AlreadySeenError:
                        continue
                    except SkipDocumentationError:
                        continue
                    current_page.extend(get_rst_action_description(level, title, action_info))

        for slug, title, page in pages:
            text = "\n".join(page) + "\n"
            filename = os.path.join(directory, "%s.rst" % slug)
            log.debug("Writing %s" % filename)
            with open(filename, "w") as fh:
                fh.write(text)

        return toc_entries

    def add_task(self, task):
        doc_hint = get_doc_hint(task)
        if "skip" in doc_hint:
            log.debug("Skipping documentation for task %s" % task.editor_id)
            return

        hierarchy = task.get_menu_action_hierarchy()
        slug = task.editor_id

        subdir = self.get_subdir(slug)
        toc_entries = self.create_task_sections(subdir, hierarchy, task.editor_id)
        template = self.get_template("task_index")

        subs = {
            "slug": slug,
            "title": get_rst_section_title(2, task.name, True),
            "toc": "\n".join([rst_toc_entry_template.format(*t) for t in toc_entries]),
            "overview": trim(task.__doc__),
        }

        text = template.format(**subs)
        filename = os.path.join(subdir, "index.rst")
        log.debug("Writing %s" % filename)
        with open(filename, "w") as fh:
            fh.write(text)

        self.sections.append((task.editor_id, task.name))

    def create_manual(self, intro=""):
        template = self.get_template("manual_index")

        subs = {
            "slug": self.title_slug,
            "title": get_rst_section_title(2, self.title, True),
            "toc": "\n".join([rst_toc_of_subdir_template.format(*t) for t in self.sections]),
            "intro": intro,
        }

        text = template.format(**subs)

        print "New manual index %s: %s" % (self.title, self.title_slug)
        log.debug("Writing index.rst")
        with open(os.path.join(self.directory, "index.rst"), "w") as fh:
            fh.write(text)

class RSTOnePageDocs(RSTSeparateMenuDocs):
    def create_task_pages(self, hierarchy, base_slug):
        toc_entries = []
        pages = []
        current_page = []
        summaries_seen = set()
        template = self.get_template("page_one_page")
        for path, action, menubar_docs in hierarchy:
            if skip_action(action):
                continue
            menu, title, level, is_action = split_path(path)
            if level > 1:
                if not is_action:  # explicit menu
                    menubar_docs = "" if menubar_docs is None else menubar_docs
                    if level == 2:  # toplevel menu item
                        slug = "%s.%s" % (base_slug, slugify(title))
                        toc_entries.append((slug, title))
                        subs = {
                            "slug": slug,
                            "title": get_rst_section_title(level, title + " Menu", False),
                            "description": menubar_docs,
                        }
                        current_page = [template.format(**subs)]
                        print "  New page for %s: %s" % (title, slug)
                        log.debug("New page for %s: %s" % (title, slug))
                        pages.append((slug, title, current_page))
                    else:
                        log.debug("Submenu %s")
                        current_page.append(get_rst_section_title(level, title, extra_docs=menubar_docs))

                else:  # menu item could be in a submenu or up a level
                    try:
                        action_info = self.get_action_info(action, menu, summaries_seen)
                    except AlreadySeenError:
                        continue
                    current_page.extend(get_rst_action_description(level, title, action_info, False))

        return toc_entries, pages

    def add_task(self, task):
        doc_hint = get_doc_hint(task)
        if "skip" in doc_hint:
            log.debug("Skipping documentation for task %s" % task.editor_id)
            return

        hierarchy = task.get_menu_action_hierarchy()
        slug = task.editor_id

        toc_entries, pages = self.create_task_pages(hierarchy, task.editor_id)
        template = self.get_template("task_index_one_page")

        subs = {
            "slug": slug,
            "title": get_rst_section_title(2, task.name, True),
            "toc": "\n".join([rst_toc_entry_template.format(*t) for t in toc_entries]),
            "overview": trim(task.__doc__),
            "pages": "\n\n".join(["\n".join(p[2]) for p in pages]),
        }

        text = template.format(**subs)
        filename = os.path.join(self.directory, "%s.rst" % slug)
        log.debug("Writing %s" % filename)
        with open(filename, "w") as fh:
            fh.write(text)

        self.sections.append((task.editor_id, task.name))

    def create_manual(self, intro=""):
        template = self.get_template("manual_index")

        subs = {
            "slug": self.title_slug,
            "title": get_rst_section_title(2, self.title, True),
            "toc": "\n".join([rst_toc_entry_template.format(*t) for t in self.sections]),
            "intro": intro,
        }

        text = template.format(**subs)

        print "New manual index %s: %s" % (self.title, self.title_slug)
        log.debug("Writing index.rst")
        with open(os.path.join(self.directory, "index.rst"), "w") as fh:
            fh.write(text)



#-------------------------------------------------------------------------------
#
#  A concrete implementation of the ITemplateDataNameItem interface that looks
#  for a specified named value in its input context or optionally in any of its
#  sub-contexts and outputs a context containing only those values that match
#  the specified name.
#
#  Written by: David C. Morrill
#
#  Date: 07/29/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" A concrete implementation of the ITemplateDataNameItem interface that looks
    for a specified named value in its input context or optionally in any of its
    sub-contexts and outputs a context containing only those values that match
    the specified name.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api \
    import Bool

from apptools.template.template_traits \
    import TStr, TBool

from apptools.template.template_choice \
    import TemplateChoice

from apptools.template.itemplate_data_context \
    import ITemplateDataContext

from any_data_name_item \
    import AnyDataNameItem

from helper \
    import parse_name, path_for

#-------------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------------

# Value used to reset to the factory settings:
ResetChoice = '--<Reset>--'

#-------------------------------------------------------------------------------
#  'ValueDataNameItem' class:
#-------------------------------------------------------------------------------

class ValueDataNameItem ( AnyDataNameItem ):
    """ A concrete implementation of the ITemplateDataNameItem interface that
        looks for a specified named value in its input context or optionally in
        any of its sub-contexts and outputs a context containing only those
        values that match the specified name.
    """

    #-- Public Traits ----------------------------------------------------------

    # The name of the value to be matched:
    name = TStr

    # (Override) Should included sub-contexts be flattened into a single
    # context?
    flatten = True

    #-- Private Traits ---------------------------------------------------------

    # The current name of the value to be matched:
    current_name = TStr

    # The current name's last component:
    current_name_last = TStr

    # Should all possible choices be included?
    all_choices = Bool( True )

    #-- AnyDataNameItem Property Implementation Overrides ----------------------

    def _get_data_name_item_choice ( self ):
        return TemplateChoice( choice_value = self.current_name )

    def _set_data_name_item_choice ( self, choice ):
        if choice.choice_value == ResetChoice:
            self.current_recursive = self.recursive
            self.current_name      = self.name
            self.all_choices       = True
        else:
            self.current_recursive = False
            self.current_name      = choice.choice_value
            self.all_choices       = False

    def _get_data_name_item_choices ( self ):
        context = self.input_data_context
        if context is None:
            return []

        if not self.all_choices:
            output_context = self.output_data_context
            if output_context is not None:
                return [ TemplateChoice( choice_value = name )
                          for name in output_context.data_context_values ]

        return ([ TemplateChoice( choice_value = ResetChoice ) ] +
                self._get_choices( context ))

    def _get_current_input_data_context ( self ):
        context = self.input_data_context
        for name in parse_name( self.current_name )[:-1]:
            if name not in context.data_contexts:
                return None

            context = context.get_data_context( name )

        return context

    #-- Abstract Method Implementations ----------------------------------------

    def filter ( self, name, value ):
        """ Returns **True** if the specified context data *name* and *value*
            should be included in the output context; and **False** otherwise.
        """
        return (name == self.current_name_last)

    #-- Trait Event Handlers ---------------------------------------------------

    def _name_changed ( self, name ):
        """ Handles the 'name' trait being changed.
        """
        self.current_name = name

    def _current_name_changed ( self, name ):
        self.current_name_last = parse_name( name )[-1]
        self.inputs_changed()

    #-- Private Methods --------------------------------------------------------

    def _get_choices ( self, context, path = '' ):
        """ Returns the list of available user settings choices for the
            specified context.
        """
        choices = [ TemplateChoice( choice_value = path_for( path, name ) )
                    for name in context.data_context_values ]

        if self.recursive:
            # Now process all of the context's sub-contexts:
            gdc = context.get_data_context
            for name in context.data_contexts:
                choices.extend( self._get_choices( gdc( name ), path_for( path,
                                    context.data_context_name ) ) )

        return choices




#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   02/14/2005
#
#------------------------------------------------------------------------------

""" Trait definition for a null-based (i.e., no UI) color.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import Trait, TraitError

#-------------------------------------------------------------------------
#  Convert a number into a wxColour object:
#-------------------------------------------------------------------------


def convert_to_color(object, name, value):
    """ Converts a number into a wxColour object.
    """
    if isinstance(value, int):
        return value & 0xFFFFFF

    elif isinstance(value, tuple):
        return (value[0] / 255., value[1] / 255., value[2] / 255.)

    raise TraitError

convert_to_color.info = ('an integer which in hex is of the form 0xRRGGBB, '
                         'where RR is red, GG is green, and BB is blue')

#-------------------------------------------------------------------------
#  Standard colors:
#-------------------------------------------------------------------------

standard_colors = {
    'aquamarine': 0x70DB93,
    'black': 0x000000,
    'blue': 0x0000FF,
    'blue violet': 0x9F5F9F,
    'brown': 0xA52A2A,
    'cadet blue': 0x5F9F9F,
    'coral': 0xFF7F00,
    'cornflower blue': 0x42426F,
    'cyan': 0x00FFFF,
    'dark grey': 0x2F2F2F,
    'dark green': 0x2F4F2F,
    'dark olive green': 0x4F4F2F,
    'dark orchid': 0x9932CC,
    'dark slate blue': 0x6B238E,
    'dark slate grey': 0x2F4F4F,
    'dark turquoise': 0x7093DB,
    'dim grey': 0x545454,
    'firebrick': 0x8E2323,
    'forest green': 0x238E23,
    'gold': 0xCC7F32,
    'goldenrod': 0xDBDB70,
    'grey': 0x808080,
    'green': 0x00FF00,
    'green yellow': 0x93DB70,
    'indian red': 0x4F2F2F,
    'khaki': 0x9F9F5F,
    'light blue': 0xBFD8D8,
    'light grey': 0xC0C0C0,
    'light steel': 0x000000,
    'blue': 0x0000FF,
    'lime green': 0x32CC32,
    'magenta': 0xFF00FF,
    'maroon': 0x8E236B,
    'medium aquamarine': 0x32CC99,
    'medium blue': 0x3232CC,
    'medium forest green': 0x6B8E23,
    'medium goldenrod': 0xEAEAAD,
    'medium orchid': 0x9370DB,
    'medium sea green': 0x426F42,
    'medium slate blue': 0x7F00FF,
    'medium spring green': 0x7FFF00,
    'medium turquoise': 0x70DBDB,
    'medium violet red': 0xDB7093,
    'midnight blue': 0x2F2F4F,
    'navy': 0x23238E,
    'orange': 0xCC3232,
    'orange red': 0xFF007F,
    'orchid': 0xDB70DB,
    'pale green': 0x8FBC8F,
    'pink': 0xBC8FEA,
    'plum': 0xEAADEA,
    'purple': 0xB000FF,
    'red': 0xFF0000,
    'salmon': 0x6F4242,
    'sea green': 0x238E6B,
    'sienna': 0x8E6B23,
    'sky blue': 0x3299CC,
    'slate blue': 0x007FFF,
    'spring green': 0x00FF7F,
    'steel blue': 0x236B8E,
    'tan': 0xDB9370,
    'thistle': 0xD8BFD8,
    'turquoise': 0xADEAEA,
    'violet': 0x4F2F4F,
    'violet red': 0xCC3299,
    'wheat': 0xD8D8BF,
    'white': 0xFFFFFF,
    'yellow': 0xFFFF00,
    'yellow green': 0x99CC32
}

#-------------------------------------------------------------------------
#  Define 'null' specific color traits:
#-------------------------------------------------------------------------

### Note: Declare the editor to be a function which returns the ColorEditor
# class from traits ui to avoid circular import issues. For backwards
# compatibility with previous Traits versions, the 'editors' folder in Traits
# project declares 'from api import *' in its __init__.py. The 'api' in turn
# can contain classes that have a Color trait which lead to this file getting
# imported. This leads to a circular import when declaring a Color trait.


def get_color_editor(*args, **traits):
    from ..api import ColorEditor
    return ColorEditor(*args, **traits)

# Color traits
NullColor = Trait('white', convert_to_color, standard_colors,
                  editor=get_color_editor)
